# Licensed under a 3-clause BSD style license - see LICENSE.rst
from gammapy.scripts.main import cli
from gammapy.utils.testing import requires_data, run_cli
from ...workflow.tests.test_workflow import get_example_config


def test_cli_workflow_config(tmp_path):
    path_config = tmp_path / "config.yaml"
    args = ["workflow", "config", f"--filename={path_config}"]
    run_cli(cli, args)
    assert path_config.exists()


@requires_data()
def test_cli_workflow_run(tmp_path):

    path_config = tmp_path / "config.yaml"
    path_datasets = tmp_path / "datasets.yaml"
    config = get_example_config("1d")
    config.datasets.background.method = "reflected"
    config.general.datasets_file = str(path_datasets)
    config.general.steps = ["data-selection"]

    config.write(path_config)
    args = [
        "workflow",
        "run",
        f"--filename={path_config}",
    ]
    run_cli(cli, args)
    assert path_datasets.exists()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from itertools import repeat
import numpy as np
import scipy.interpolate
import scipy.ndimage as ndi
import scipy.signal
import astropy.units as u
from astropy.convolution import Tophat2DKernel
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.nddata import block_reduce
from regions import PixCoord, PointPixelRegion, PointSkyRegion, SkyRegion
import matplotlib.colors as mpcolors
import matplotlib.pyplot as plt
import gammapy.utils.parallel as parallel
from gammapy.utils.interpolation import ScaledRegularGridInterpolator
from gammapy.utils.units import unit_from_fits_image_hdu
from gammapy.visualization.utils import add_colorbar
from ..geom import pix_tuple_to_idx
from ..utils import INVALID_INDEX
from .core import WcsMap
from .geom import WcsGeom

__all__ = ["WcsNDMap"]

log = logging.getLogger(__name__)


C_MAP_MASK = mpcolors.ListedColormap(["black", "white"], name="mask")


class WcsNDMap(WcsMap):
    """WCS map with any number of non-spatial dimensions.

    This class uses an ND numpy array to store map values. For maps with
    non-spatial dimensions and variable pixel size it will allocate an
    array with dimensions commensurate with the largest image plane.

    Parameters
    ----------
    geom : `~gammapy.maps.WcsGeom`
        WCS geometry object.
    data : `~numpy.ndarray`
        Data array. If none then an empty array will be allocated.
    dtype : str, optional
        Data type, default is float32
    meta : `dict`
        Dictionary to store meta data.
    unit : str or `~astropy.units.Unit`
        The map unit
    """

    def __init__(self, geom, data=None, dtype="float32", meta=None, unit=""):
        # TODO: Figure out how to mask pixels for integer data types

        data_shape = geom.data_shape

        if data is None:
            data = self._make_default_data(geom, data_shape, dtype)

        super().__init__(geom, data, meta, unit)

    @staticmethod
    def _make_default_data(geom, shape_np, dtype):
        # Check whether corners of each image plane are valid

        data = np.zeros(shape_np, dtype=dtype)

        if not geom.is_regular or geom.is_allsky:
            coords = geom.get_coord()
            is_nan = np.isnan(coords.lon)
            data[is_nan] = np.nan

        return data

    @classmethod
    def from_hdu(cls, hdu, hdu_bands=None, format=None):
        """Make a WcsNDMap object from a FITS HDU.

        Parameters
        ----------
        hdu : `~astropy.io.fits.BinTableHDU` or `~astropy.io.fits.ImageHDU`
            The map FITS HDU.
        hdu_bands : `~astropy.io.fits.BinTableHDU`
            The BANDS table HDU.
        format : {'gadf', 'fgst-ccube','fgst-template'}
            FITS format convention.

        Returns
        -------
        map : `WcsNDMap`
            WCS map.
        """
        geom = WcsGeom.from_header(hdu.header, hdu_bands, format=format)
        shape = geom.axes.shape
        shape_wcs = tuple([np.max(geom.npix[0]), np.max(geom.npix[1])])

        meta = cls._get_meta_from_header(hdu.header)
        unit = unit_from_fits_image_hdu(hdu.header)

        # TODO: Should we support extracting slices?
        if isinstance(hdu, fits.BinTableHDU):
            map_out = cls(geom, meta=meta, unit=unit)
            pix = hdu.data.field("PIX")
            pix = np.unravel_index(pix, shape_wcs[::-1])
            vals = hdu.data.field("VALUE")
            if "CHANNEL" in hdu.data.columns.names and shape:
                chan = hdu.data.field("CHANNEL")
                chan = np.unravel_index(chan, shape[::-1])
                idx = chan + pix
            else:
                idx = pix

            map_out.set_by_idx(idx[::-1], vals)
        else:
            if any(x in hdu.name.lower() for x in ["mask", "is_ul", "success"]):
                data = hdu.data.astype(bool)
            else:
                data = hdu.data

            map_out = cls(geom=geom, meta=meta, data=data, unit=unit)

        return map_out

    def get_by_idx(self, idx):
        idx = pix_tuple_to_idx(idx)
        return self.data.T[idx]

    def interp_by_coord(
        self, coords, method="linear", fill_value=None, values_scale="lin"
    ):
        """Interpolate map values at the given map coordinates.

        Parameters
        ----------
        coords : tuple, dict or `~gammapy.maps.MapCoord`
            Coordinate arrays for each dimension of the map. Tuple
            should be ordered as (lon, lat, x_0, ..., x_n) where x_i
            are coordinates for non-spatial dimensions of the map.
            "lon" and "lat" are optional and will be taken at the center
            of the region by default.
        method : {"linear", "nearest"}
            Method to interpolate data values. By default linear
            interpolation is performed.
        fill_value : None or float value
            The value to use for points outside of the interpolation domain.
            If None, values outside the domain are extrapolated.
        values_scale : {"lin", "log", "sqrt"}
            Optional value scaling. Default is "lin".

        Returns
        -------
        vals : `~numpy.ndarray`
            Interpolated pixel values.
        """
        if self.geom.is_regular:
            pix = self.geom.coord_to_pix(coords)
            return self.interp_by_pix(
                pix, method=method, fill_value=fill_value, values_scale=values_scale
            )
        else:
            return self._interp_by_coord_griddata(coords, method=method)

    def interp_by_pix(self, pix, method="linear", fill_value=None, values_scale="lin"):
        if not self.geom.is_regular:
            raise ValueError("interp_by_pix only supported for regular geom.")

        grid_pix = [np.arange(n, dtype=float) for n in self.data.shape[::-1]]

        if np.any(np.isfinite(self.data)):
            data = self.data.copy().T
            data[~np.isfinite(data)] = 0.0
        else:
            data = self.data.T

        fn = ScaledRegularGridInterpolator(
            grid_pix,
            data,
            fill_value=None,
            bounds_error=False,
            method=method,
            values_scale=values_scale,
        )
        interp_data = fn(tuple(pix), clip=False)

        if fill_value is not None:
            idxs = self.geom.pix_to_idx(pix, clip=False)
            invalid = np.broadcast_arrays(*[idx == -1 for idx in idxs])
            mask = np.any(invalid, axis=0)
            if not interp_data.shape:
                mask = mask.squeeze()
            interp_data[mask] = fill_value
            interp_data[~np.isfinite(interp_data)] = fill_value

        return interp_data

    def _interp_by_coord_griddata(self, coords, method="linear"):
        grid_coords = self.geom.get_coord()

        data = self.data[np.isfinite(self.data)]
        vals = scipy.interpolate.griddata(
            tuple(grid_coords.flat), data, tuple(coords), method=method
        )

        m = ~np.isfinite(vals)
        if np.any(m):
            vals_fill = scipy.interpolate.griddata(
                tuple(grid_coords.flat),
                data,
                tuple([c[m] for c in coords]),
                method="nearest",
            )
            vals[m] = vals_fill

        return vals

    def _resample_by_idx(self, idx, weights=None, preserve_counts=False, smooth=False):
        idx = pix_tuple_to_idx(idx)
        msk = np.all(np.stack([t != INVALID_INDEX.int for t in idx]), axis=0)
        idx = [t[msk] for t in idx]

        if weights is not None:
            if isinstance(weights, u.Quantity):
                weights = weights.to_value(self.unit)
            weights = weights[msk]

        idx = np.ravel_multi_index(idx, self.data.T.shape)
        idx, idx_inv = np.unique(idx, return_inverse=True)

        weights = np.bincount(idx_inv, weights=weights).astype(self.data.dtype)
        bincount = np.bincount(idx_inv).astype(self.data.dtype)

        if smooth:
            weight_sum = np.nansum(weights)
            weights /= bincount
            if preserve_counts:
                factor = weight_sum / np.nansum(weights)
                weights *= np.nan_to_num(factor, nan=0.0, posinf=0.0, neginf=0.0)
        elif not preserve_counts:
            weights /= bincount

        self.data.T.flat[idx] += weights

    def fill_by_idx(self, idx, weights=None):
        return self._resample_by_idx(idx, weights=weights, preserve_counts=True)

    def set_by_idx(self, idx, vals):
        idx = pix_tuple_to_idx(idx)
        self.data.T[idx] = vals

    def _pad_spatial(
        self, pad_width, axis_name=None, mode="constant", cval=0, method="linear"
    ):
        if axis_name is None:
            if np.isscalar(pad_width):
                pad_width = (pad_width, pad_width)

            if len(pad_width) == 2:
                pad_width += (0,) * (self.geom.ndim - 2)

            geom = self.geom._pad_spatial(pad_width[:2])
            if self.geom.is_regular and mode != "interp":
                return self._pad_np(geom, pad_width, mode, cval)
            else:
                return self._pad_coadd(geom, pad_width, mode, cval, method)

    def _pad_np(self, geom, pad_width, mode, cval):
        """Pad a map using `~numpy.pad`.

        This method only works for regular geometries but should be more
        efficient when working with large maps.
        """
        kwargs = {}
        if mode == "constant":
            kwargs["constant_values"] = cval

        pad_width = [(t, t) for t in pad_width]
        data = np.pad(self.data, pad_width[::-1], mode, **kwargs)
        return self._init_copy(geom=geom, data=data)

    def _pad_coadd(self, geom, pad_width, mode, cval, method):
        """Pad a map manually by co-adding the original map with the new map."""
        idx_in = self.geom.get_idx(flat=True)
        idx_in = tuple([t + w for t, w in zip(idx_in, pad_width)])[::-1]
        idx_out = geom.get_idx(flat=True)[::-1]
        map_out = self._init_copy(geom=geom, data=None)
        map_out.coadd(self)

        if mode == "constant":
            pad_msk = np.zeros_like(map_out.data, dtype=bool)
            pad_msk[idx_out] = True
            pad_msk[idx_in] = False
            map_out.data[pad_msk] = cval
        elif mode == "interp":
            coords = geom.pix_to_coord(idx_out[::-1])
            m = self.geom.contains(coords)
            coords = tuple([c[~m] for c in coords])
            vals = self.interp_by_coord(coords, method=method)
            map_out.set_by_coord(coords, vals)
        else:
            raise ValueError(f"Invalid mode: {mode!r}")

        return map_out

    def crop(self, crop_width):
        if np.isscalar(crop_width):
            crop_width = (crop_width, crop_width)

        geom = self.geom.crop(crop_width)
        if self.geom.is_regular:
            slices = [slice(None)] * len(self.geom.axes)
            slices += [
                slice(crop_width[1], int(self.geom.npix[1][0] - crop_width[1])),
                slice(crop_width[0], int(self.geom.npix[0][0] - crop_width[0])),
            ]
            data = self.data[tuple(slices)]
            map_out = self._init_copy(geom=geom, data=data)
        else:
            # FIXME: This could be done more efficiently by
            # constructing the appropriate slices for each image plane
            map_out = self._init_copy(geom=geom, data=None)
            map_out.coadd(self)

        return map_out

    def upsample(self, factor, order=0, preserve_counts=True, axis_name=None):
        if factor == 1 or factor is None:
            return self

        geom = self.geom.upsample(factor, axis_name=axis_name)
        idx = geom.get_idx()

        if axis_name is None:
            pix = (
                (idx[0] - 0.5 * (factor - 1)) / factor,
                (idx[1] - 0.5 * (factor - 1)) / factor,
            ) + idx[2:]
        else:
            pix = list(idx)
            idx_ax = self.geom.axes_names.index(axis_name)
            pix[idx_ax] = (pix[idx_ax] - 0.5 * (factor - 1)) / factor

        if preserve_counts:
            data = self.data / self.geom.bin_volume().value
        else:
            data = self.data

        data = ndi.map_coordinates(data.T, tuple(pix), order=order, mode="nearest")

        if preserve_counts:
            data *= geom.bin_volume().value

        return self._init_copy(geom=geom, data=data.astype(self.data.dtype))

    def downsample(self, factor, preserve_counts=True, axis_name=None, weights=None):
        if factor == 1 or factor is None:
            return self

        geom = self.geom.downsample(factor, axis_name=axis_name)

        if axis_name is None:
            block_size = (1,) * len(self.geom.axes) + (factor, factor)
        else:
            block_size = [1] * self.data.ndim
            idx = self.geom.axes.index_data(axis_name)
            block_size[idx] = factor

        if weights is None:
            weights = np.ones_like(self.data)
        elif not isinstance(weights, np.ndarray):
            weights = weights.data

        data = block_reduce(self.data * weights, tuple(block_size), func=np.nansum)

        if not preserve_counts:
            weight_sum = block_reduce(weights, tuple(block_size), func=np.nansum)
            data = np.divide(
                data,
                weight_sum,
                out=np.zeros_like(data, dtype="float64"),
                where=(weight_sum != 0),
            )

        return self._init_copy(geom=geom, data=data.astype(self.data.dtype))

    def plot(
        self,
        ax=None,
        fig=None,
        add_cbar=False,
        stretch="linear",
        axes_loc=None,
        kwargs_colorbar=None,
        **kwargs,
    ):
        """
        Plot image on matplotlib WCS axes.

        Parameters
        ----------
        ax : `~astropy.visualization.wcsaxes.WCSAxes`, optional
            WCS axis object to plot on. Default is None.
        fig : `~matplotlib.figure.Figure`, optional
            Figure object. Default is None.
        add_cbar : bool, optional
            Add color bar. Default is False.
        stretch : str, optional
            Passed to `astropy.visualization.simple_norm`.
             Default is "linear".
        axes_loc : dict, optional
            Keyword arguments passed to `~mpl_toolkits.axes_grid1.axes_divider.AxesDivider.append_axes`.
        kwargs_colorbar : dict, optional
            Keyword arguments passed to `~matplotlib.pyplot.colorbar`.
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.pyplot.imshow`.

        Returns
        -------
        ax : `~astropy.visualization.wcsaxes.WCSAxes`
            WCS axes object.
        """
        from astropy.visualization import simple_norm

        if not self.geom.is_flat:
            raise TypeError("Use .plot_interactive() for Map dimension > 2")

        ax = self._plot_default_axes(ax=ax)

        if fig is None:
            fig = plt.gcf()

        if self.geom.is_image:
            data = self.data.astype(float)
        else:
            axis = tuple(np.arange(len(self.geom.axes)))
            data = np.squeeze(self.data, axis=axis).astype(float)

        kwargs.setdefault("interpolation", "nearest")
        kwargs.setdefault("origin", "lower")
        kwargs.setdefault("cmap", "afmhot")

        kwargs_colorbar = kwargs_colorbar or {}

        mask = np.isfinite(data)

        if self.is_mask:
            kwargs.setdefault("vmin", 0)
            kwargs.setdefault("vmax", 1)
            kwargs["cmap"] = C_MAP_MASK

        if mask.any():
            min_cut, max_cut = kwargs.pop("vmin", None), kwargs.pop("vmax", None)
            try:
                norm = simple_norm(data[mask], stretch, vmin=min_cut, vmax=max_cut)
            except TypeError:
                # astropy <6.1
                norm = simple_norm(
                    data[mask], stretch, min_cut=min_cut, max_cut=max_cut
                )
            kwargs.setdefault("norm", norm)

        im = ax.imshow(data, **kwargs)

        if add_cbar:
            label = str(self.unit)
            kwargs_colorbar.setdefault("label", label)
            add_colorbar(im, ax=ax, axes_loc=axes_loc, **kwargs_colorbar)

        if self.geom.is_allsky:
            ax = self._plot_format_allsky(ax)
        else:
            ax = self._plot_format(ax)

        # without this the axis limits are changed when calling scatter
        ax.autoscale(enable=False)
        return ax

    def plot_mask(self, ax=None, **kwargs):
        """Plot the mask as a shaded area.

        Parameters
        ----------
        ax : `~astropy.visualization.wcsaxes.WCSAxes`, optional
            WCS axis object to plot on. Default is None.

        **kwargs : dict
            Keyword arguments passed to `~matplotlib.pyplot.contourf`.

        Returns
        -------
        ax : `~astropy.visualization.wcsaxes.WCSAxes`, optional
            WCS axis object to plot on.
        """
        if not self.geom.is_flat:
            raise TypeError("Use .plot_interactive() for Map dimension > 2")

        if not self.is_mask:
            raise ValueError(
                "`.plot_mask()` only supports maps containing boolean values."
            )

        ax = self._plot_default_axes(ax=ax)

        kwargs.setdefault("alpha", 0.5)
        kwargs.setdefault("colors", "w")

        data = np.squeeze(self.data).astype(float)

        ax.contourf(data, levels=[0, 0.5], **kwargs)

        if self.geom.is_allsky:
            ax = self._plot_format_allsky(ax)
        else:
            ax = self._plot_format(ax)

        # without this the axis limits are changed when calling scatter
        ax.autoscale(enable=False)
        return ax

    def _plot_default_axes(self, ax):
        from astropy.visualization.wcsaxes.frame import EllipticalFrame

        if ax is None:
            fig = plt.gcf()
            if self.geom.projection in ["AIT"]:
                ax = fig.add_subplot(
                    1, 1, 1, projection=self.geom.wcs, frame_class=EllipticalFrame
                )
            else:
                ax = fig.add_subplot(1, 1, 1, projection=self.geom.wcs)

        return ax

    @staticmethod
    def _plot_format(ax):
        try:
            ax.coords["glon"].set_axislabel("Galactic Longitude")
            ax.coords["glat"].set_axislabel("Galactic Latitude")
        except KeyError:
            ax.coords["ra"].set_axislabel("Right Ascension")
            ax.coords["dec"].set_axislabel("Declination")
        except AttributeError:
            log.info("Can't set coordinate axes. No WCS information available.")
        return ax

    def _plot_format_allsky(self, ax):
        # Remove frame
        ax.coords.frame.set_linewidth(0)

        # Set plot axis limits
        xmin, _ = self.geom.to_image().coord_to_pix({"lon": 180, "lat": 0})
        xmax, _ = self.geom.to_image().coord_to_pix({"lon": -180, "lat": 0})

        _, ymin = self.geom.to_image().coord_to_pix({"lon": 0, "lat": -90})
        _, ymax = self.geom.to_image().coord_to_pix({"lon": 0, "lat": 90})

        ax.set_xlim(xmin[0], xmax[0])
        ax.set_ylim(ymin[0], ymax[0])

        ax.text(0, ymax[0], self.geom.frame + " coords")

        # Grid and ticks
        glon_spacing, glat_spacing = 45, 15
        lon, lat = ax.coords
        lon.set_ticks(spacing=glon_spacing * u.deg, color="w", alpha=0.8)
        lat.set_ticks(spacing=glat_spacing * u.deg)
        lon.set_ticks_visible(False)

        lon.set_major_formatter("d")
        lat.set_major_formatter("d")

        lon.set_ticklabel(color="w", alpha=0.8)
        lon.grid(alpha=0.2, linestyle="solid", color="w")
        lat.grid(alpha=0.2, linestyle="solid", color="w")
        return ax

    def cutout_and_mask_region(self, region=None):
        """Compute cutout and mask for a given region of the map.

        The function will estimate the minimal size of the cutout, which encloses
        the region.

        Parameters
        ----------
        region: `~regions.Region`, optional
             Extended region. Default is None.

        Returns
        -------
        cutout, mask : tuple of `WcsNDMap`
            Cutout and mask map.
        """
        from gammapy.maps import RegionGeom

        if region is None:
            region = self.geom.footprint_rectangle_sky_region

        geom = RegionGeom.from_regions(regions=region, wcs=self.geom.wcs)
        cutout = self.cutout(position=geom.center_skydir, width=geom.width)

        mask = cutout.geom.to_image().region_mask([region])
        return self.__class__(data=cutout.data, geom=cutout.geom, unit=self.unit), mask

    def to_region_nd_map(
        self, region=None, func=np.nansum, weights=None, method="nearest"
    ):
        """Get region ND map in a given region.

        By default, the whole map region is considered.

        Parameters
        ----------
        region: `~regions.Region` or `~astropy.coordinates.SkyCoord`, optional
             Region. Default is None.
        func : numpy.func, optional
            Function to reduce the data. Default is np.nansum.
            For boolean Map, use np.any or np.all.
        weights : `WcsNDMap`, optional
            Array to be used as weights.
            Weights are multiplied with the input data and should be defined accordingly, in case a weighted average is desired.
            If a mask is given, the mask is applied to the input map. The geometry must be equivalent.
            Default is None.
        method : {"nearest", "linear"}, optional
            How to interpolate if a position is given.
            Default is "nearest".

        Returns
        -------
        spectrum : `~gammapy.maps.RegionNDMap`
            Spectrum in the given region.
        """
        from gammapy.maps import RegionGeom, RegionNDMap

        if region is None:
            region = self.geom.footprint_rectangle_sky_region

        if weights is not None:
            if not self.geom == weights.geom:
                raise ValueError("Incompatible spatial geoms between map and weights")

        geom = RegionGeom.from_regions(
            regions=region, axes=self.geom.axes, wcs=self.geom.wcs
        )

        if geom.is_all_point_sky_regions:
            coords = geom.get_coord()
            data = self.interp_by_coord(coords=coords, method=method)

            if weights is not None:
                data *= weights.interp_by_coord(coords=coords, method=method)
            # Casting needed as interp_by_coord transforms boolean
            data = data.astype(self.data.dtype)
        else:
            cutout, cutout_mask = self.cutout_and_mask_region(region=region)
            mask = cutout_mask.interp_to_geom(cutout.geom, method="nearest")

            if weights is not None:
                weights_cutout = weights.cutout(
                    position=geom.center_skydir, width=geom.width
                )
                if weights_cutout.is_mask:
                    mask.data = np.logical_and(mask.data, weights_cutout.data)
                else:
                    cutout.data *= weights_cutout.data

            data = np.empty(cutout.data.shape[:-2], dtype=self.data.dtype)

            for i, val in np.ndenumerate(data):
                data[i] = func(cutout.data[i][mask.data[i]]).astype(self.data.dtype)

        return RegionNDMap(geom=geom, data=data, unit=self.unit, meta=self.meta.copy())

    def to_region_nd_map_histogram(
        self, region=None, bins_axis=None, nbin=100, density=False
    ):
        """Convert map into region map by histogramming.

        By default, it creates a linearly spaced axis with 100 bins between
        (-max(abs(data)), max(abs(data))) within the given region.

        Parameters
        ----------
        region: `~regions.Region`, optional
            Region to histogram over. Default is None.
        bins_axis : `MapAxis`, optional
            Binning of the histogram. Default is None.
        nbin : int, optional
            Number of bins to use if no bins_axis is given.
            Default is 100.
        density : bool, optional
            Normalize integral of the histogram to 1.
            Default is False.


        Examples
        --------
        This is how to use the method to create energy dependent histograms:

        ::

            from gammapy.maps import MapAxis, Map
            import numpy as np

            random_state = np.random.RandomState(seed=0)

            energy_axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=3)

            data = Map.create(axes=[energy_axis], width=10, unit="cm2 s-1", binsz=0.02)
            data.data = random_state.normal(
                size=data.data.shape, loc=0, scale=np.array([1.0, 2.0, 3.0]).reshape((-1, 1, 1))
            )

            hist = data.to_region_nd_map_histogram()
            hist.plot(axis_name="bins")


        Returns
        -------
        region_map : `RegionNDMap`
            Region map with histogram.

        """
        from gammapy.maps import MapAxis, RegionGeom, RegionNDMap

        if isinstance(region, (PointSkyRegion, SkyCoord)):
            raise ValueError("Histogram method not supported for point regions")

        cutout, mask = self.cutout_and_mask_region(region=region)
        idx_y, idx_x = np.where(mask)
        quantity = cutout.quantity[..., idx_y, idx_x]

        value = np.abs(quantity).max()

        if bins_axis is None:
            bins_axis = MapAxis.from_bounds(
                -value,
                value,
                nbin=nbin,
                interp="lin",
                unit=self.unit,
                name="bins",
            )

        if not bins_axis.unit.is_equivalent(self.unit):
            raise ValueError("Unit of bins_axis must be equivalent to unit of map.")

        axes = [bins_axis] + list(self.geom.axes)
        geom_hist = RegionGeom(region=region, axes=axes, wcs=self.geom.wcs)

        # This is likely not the most efficient way to do this
        data = np.apply_along_axis(
            lambda a: np.histogram(a, bins=bins_axis.edges.value, density=density)[0],
            axis=-1,
            arr=quantity.to_value(bins_axis.unit),
        )

        unit = 1.0 / bins_axis.unit if density else ""

        return RegionNDMap.from_geom(geom=geom_hist, data=data, unit=unit)

    def mask_contains_region(self, region):
        """Check if input region is contained in a boolean mask map.

        Parameters
        ----------
        region: `~regions.SkyRegion` or `~regions.PixRegion`
             Region or list of Regions (pixel or sky regions accepted).

        Returns
        -------
        contained : bool
            Whether region is contained in the mask.
        """
        if not self.is_mask:
            raise ValueError("mask_contains_region is only supported for boolean masks")

        if not self.geom.is_image:
            raise ValueError("Method only supported for 2D images")

        if isinstance(region, SkyRegion):
            region = region.to_pixel(self.geom.wcs)

        if isinstance(region, PointPixelRegion):
            lon, lat = region.center.x, region.center.y
            contains = self.get_by_pix((lon, lat))
        else:
            idx = self.geom.get_idx()
            coords_pix = PixCoord(idx[0][self.data], idx[1][self.data])
            contains = region.contains(coords_pix)

        return np.any(contains)

    def binary_erode(self, width, kernel="disk", use_fft=True):
        """Binary erosion of boolean mask removing a given margin.

        Parameters
        ----------
        width : `~astropy.units.Quantity`, str or float
            If a float is given it interpreted as width in pixels. If an (angular)
            quantity is given it converted to pixels using ``geom.wcs.wcs.cdelt``.
            The width corresponds to radius in case of a disk kernel, and
            the side length in case of a box kernel.
        kernel : {'disk', 'box'}, optional
            Kernel shape. Default is "disk".
        use_fft : bool, optional
            Use `scipy.signal.fftconvolve` if True. Otherwise, use
            `scipy.ndimage.binary_erosion`.
            Default is True.


        Returns
        -------
        map : `WcsNDMap`
            Eroded mask map.

        """
        if not self.is_mask:
            raise ValueError("Binary operations only supported for boolean masks")

        structure = self.geom.binary_structure(width=width, kernel=kernel)

        if use_fft:
            return self.convolve(structure.squeeze(), method="fft") > (
                structure.sum() - 1
            )

        data = ndi.binary_erosion(self.data, structure=structure)
        return self._init_copy(data=data)

    def binary_dilate(self, width, kernel="disk", use_fft=True):
        """Binary dilation of boolean mask adding a given margin.

        Parameters
        ----------
        width : tuple of `~astropy.units.Quantity`
            Angular sizes of the margin in (lon, lat) in that specific order.
            If only one value is passed, the same margin is applied in (lon, lat).
        kernel : {'disk', 'box'}, optional
            Kernel shape. Default is "disk".
        use_fft : bool, optional
            Use `scipy.signal.fftconvolve` if True. Otherwise, use
            `scipy.ndimage.binary_dilation`.
            Default is True.

        Returns
        -------
        map : `WcsNDMap`
            Dilated mask map.
        """
        if not self.is_mask:
            raise ValueError("Binary operations only supported for boolean masks")

        structure = self.geom.binary_structure(width=width, kernel=kernel)

        if use_fft:
            return self.convolve(structure.squeeze(), method="fft") > 1

        data = ndi.binary_dilation(self.data, structure=structure)
        return self._init_copy(data=data)

    def convolve(self, kernel, method="fft", mode="same"):
        """Convolve map with a kernel.

        If the kernel is two-dimensional, it is applied to all image planes likewise.
        If the kernel is higher dimensional, it should either match the map in the number of
        dimensions or the map must be an image (no non-spatial axes). In that case, the
        corresponding kernel is selected and applied to every image plane or to the single
        input image respectively.

        Parameters
        ----------
        kernel : `~gammapy.irf.PSFKernel` or `numpy.ndarray`
            Convolution kernel.
        method : str, optional
            The method used by `~scipy.signal.convolve`.
            Default is 'fft'.
        mode : str, optional
            The convolution mode used by `~scipy.signal.convolve`.
            Default is 'same'.

        Returns
        -------
        map : `WcsNDMap`
            Convolved map.
        """
        from gammapy.irf import PSFKernel

        if self.geom.is_image and not isinstance(kernel, PSFKernel):
            if kernel.ndim > 2:
                raise ValueError(
                    "Image convolution with 3D kernel requires a PSFKernel object"
                )

        geom = self.geom.copy()

        if isinstance(kernel, PSFKernel):
            kmap = kernel.psf_kernel_map
            if not np.allclose(
                self.geom.pixel_scales.deg, kmap.geom.pixel_scales.deg, rtol=1e-5
            ):
                raise ValueError("Pixel size of kernel and map not compatible.")
            kernel = kmap.data.astype(np.float32)
            if self.geom.is_image:
                geom = geom.to_cube(kmap.geom.axes)

        if mode == "full":
            pad_width = [0.5 * (width - 1) for width in kernel.shape[-2:]]
            geom = geom.pad(pad_width, axis_name=None)
        elif mode == "valid":
            raise NotImplementedError(
                "WcsNDMap.convolve: mode='valid' is not supported."
            )

        shape_axes_kernel = kernel.shape[slice(0, -2)]

        if len(shape_axes_kernel) > 0:
            if not geom.shape_axes == shape_axes_kernel:
                raise ValueError(
                    f"Incompatible shape between data {geom.shape_axes}"
                    " and kernel {shape_axes_kernel}"
                )

        if self.geom.is_image and kernel.ndim == 3:
            indexes = range(kernel.shape[0])
            images = repeat(self.data.astype(np.float32))
        else:
            indexes = list(self.iter_by_image_index())
            images = (self.data[idx] for idx in indexes)
        kernels = (
            kernel[Ellipsis] if kernel.ndim == 2 else kernel[idx] for idx in indexes
        )

        convolved = parallel.run_multiprocessing(
            self._convolve,
            zip(
                images,
                kernels,
                repeat(method),
                repeat(mode),
            ),
            task_name="Convolution",
        )
        data = np.empty(geom.data_shape, dtype=np.float32)
        for idx_res, idx in enumerate(indexes):
            data[idx] = convolved[idx_res]
        return self._init_copy(data=data, geom=geom)

    @staticmethod
    def _convolve(image, kernel, method, mode):
        """Convolve using `~scipy.signal.convolve` without kwargs for parallel evaluation."""
        return scipy.signal.convolve(image, kernel, method=method, mode=mode)

    def smooth(self, width, kernel="gauss", **kwargs):
        """Smooth the map.

        Iterates over 2D image planes, processing one at a time.

        Parameters
        ----------
        width : `~astropy.units.Quantity`, str or float
            Smoothing width given as quantity or float. If a float is given it
            interpreted as smoothing width in pixels. If an (angular) quantity
            is given it converted to pixels using ``geom.wcs.wcs.cdelt``.
            It corresponds to the standard deviation in case of a Gaussian kernel,
            the radius in case of a disk kernel, and the side length in case
            of a box kernel.
        kernel : {'gauss', 'disk', 'box'}, optional
            Kernel shape. Default is "gauss".
        kwargs : dict
            Keyword arguments passed to `~ndi.uniform_filter`
            ('box'), `~ndi.gaussian_filter` ('gauss') or
            `~ndi.convolve` ('disk').

        Returns
        -------
        image : `WcsNDMap`
            Smoothed image (a copy, the original object is unchanged).
        """
        if isinstance(width, (u.Quantity, str)):
            width = u.Quantity(width) / self.geom.pixel_scales.mean()
            width = width.to_value("")

        smoothed_data = np.empty(self.data.shape, dtype=float)

        for img, idx in self.iter_by_image_data():
            img = img.astype(float)
            if kernel == "gauss":
                data = ndi.gaussian_filter(img, width, **kwargs)
            elif kernel == "disk":
                disk = Tophat2DKernel(width)
                disk.normalize("integral")
                data = ndi.convolve(img, disk.array, **kwargs)
            elif kernel == "box":
                data = ndi.uniform_filter(img, width, **kwargs)
            else:
                raise ValueError(f"Invalid kernel: {kernel!r}")
            smoothed_data[idx] = data

        return self._init_copy(data=smoothed_data)

    def cutout(self, position, width, mode="trim", odd_npix=False, min_npix=1):
        """
        Create a cutout around a given position.

        Parameters
        ----------
        position : `~astropy.coordinates.SkyCoord`
            Center position of the cutout region.
        width : tuple of `~astropy.coordinates.Angle`
            Angular sizes of the region in (lon, lat) in that specific order.
            If only one value is passed, a square region is extracted.
        mode : {'trim', 'partial', 'strict'}, optional
            Mode option for Cutout2D, for details see `~astropy.nddata.utils.Cutout2D`.
            Default is "trim".
        odd_npix : bool, optional
            Force width to odd number of pixels.
            Default is False.
        min_npix : bool, optional
            Force width to a minimmum number of pixels.
            Default is 1.

        Returns
        -------
        cutout : `~gammapy.maps.WcsNDMap`
            Cutout map.
        """
        geom_cutout = self.geom.cutout(
            position=position,
            width=width,
            mode=mode,
            odd_npix=odd_npix,
            min_npix=min_npix,
        )
        cutout_info = geom_cutout.cutout_slices(self.geom, mode=mode)

        slices = cutout_info["parent-slices"]
        parent_slices = Ellipsis, slices[0], slices[1]

        slices = cutout_info["cutout-slices"]
        cutout_slices = Ellipsis, slices[0], slices[1]

        data = np.zeros(shape=geom_cutout.data_shape, dtype=self.data.dtype)
        data[cutout_slices] = self.data[parent_slices]

        return self._init_copy(geom=geom_cutout, data=data)

    def _cutout_view(self, position, width, odd_npix=False):
        """
        Create a cutout around a given position without copy of the data.

        Parameters
        ----------
        position : `~astropy.coordinates.SkyCoord`
            Center position of the cutout region.
        width : tuple of `~astropy.coordinates.Angle`
            Angular sizes of the region in (lon, lat) in that specific order.
            If only one value is passed, a square region is extracted.
        odd_npix : bool, optional
            Force width to odd number of pixels.
            Default is False.

        Returns
        -------
        cutout : `~gammapy.maps.WcsNDMap`
            Cutout map.
        """
        geom_cutout = self.geom.cutout(
            position=position, width=width, mode="trim", odd_npix=odd_npix
        )
        cutout_info = geom_cutout.cutout_slices(self.geom, mode="trim")

        slices = cutout_info["parent-slices"]
        parent_slices = Ellipsis, slices[0], slices[1]

        return self.__class__.from_geom(
            geom=geom_cutout, data=self.quantity[parent_slices]
        )

    def stack(self, other, weights=None, nan_to_num=True):
        """Stack cutout into map.

        Parameters
        ----------
        other : `WcsNDMap`
            Other map to stack.
        weights : `WcsNDMap`, optional
            Array to be used as weights. The spatial geometry must be equivalent
            to `other` and additional axes must be broadcastable.
            Default is None.
        nan_to_num: bool, optional
            Non-finite values are replaced by zero if True.
            Default is True.

        """
        if self.geom == other.geom:
            parent_slices, cutout_slices = None, None
        elif self.geom.is_aligned(other.geom):
            cutout_slices = other.geom.cutout_slices(self.geom)

            slices = cutout_slices["parent-slices"]
            parent_slices = Ellipsis, slices[0], slices[1]

            slices = cutout_slices["cutout-slices"]
            cutout_slices = Ellipsis, slices[0], slices[1]
        else:
            raise ValueError(
                "Can only stack equivalent maps or cutout of the same map."
            )

        data = other.quantity[cutout_slices].to_value(self.unit)
        if nan_to_num:
            not_finite = ~np.isfinite(data)
            if np.any(not_finite):
                data = data.copy()
                data[not_finite] = 0
        if weights is not None:
            if not other.geom.to_image() == weights.geom.to_image():
                raise ValueError("Incompatible spatial geoms between map and weights")
            data = data * weights.data[cutout_slices]
        self.data[parent_slices] += data



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import logging
import numpy as np
from numpy.testing import assert_allclose, assert_equal
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.io import fits
from astropy.time import Time
from regions import CircleSkyRegion, PointSkyRegion
from gammapy.maps import Map, MapAxis, TimeMapAxis, WcsGeom
from gammapy.maps.utils import _check_binsz, _check_width
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data, requires_dependency
import gammapy.utils.cache as cache

axes1 = [MapAxis(np.logspace(0.0, 3.0, 3), interp="log", name="energy")]
axes2 = [
    MapAxis(np.logspace(0.0, 3.0, 3), interp="log", name="energy"),
    MapAxis(np.logspace(1.0, 3.0, 4), interp="lin"),
]
skydir = SkyCoord(110.0, 75.0, unit="deg", frame="icrs")

wcs_allsky_test_geoms = [
    (None, 10.0, "galactic", "AIT", skydir, None),
    (None, 10.0, "galactic", "AIT", skydir, axes1),
    (None, [10.0, 20.0], "galactic", "AIT", skydir, axes1),
    (None, 10.0, "galactic", "AIT", skydir, axes2),
    (None, [[10.0, 20.0, 30.0], [10.0, 20.0, 30.0]], "galactic", "AIT", skydir, axes2),
]

wcs_partialsky_test_geoms = [
    (10, 0.1, "galactic", "AIT", skydir, None),
    (10, 0.1, "galactic", "AIT", skydir, axes1),
    (10, [0.1, 0.2], "galactic", "AIT", skydir, axes1),
]

wcs_test_geoms = wcs_allsky_test_geoms + wcs_partialsky_test_geoms


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skydir", "axes"), wcs_test_geoms
)
def test_wcsgeom_init(npix, binsz, frame, proj, skydir, axes):
    WcsGeom.create(
        npix=npix, binsz=binsz, skydir=skydir, proj=proj, frame=frame, axes=axes
    )


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skydir", "axes"), wcs_test_geoms
)
def test_wcsgeom_get_pix(npix, binsz, frame, proj, skydir, axes):
    geom = WcsGeom.create(
        npix=npix, binsz=binsz, skydir=skydir, proj=proj, frame=frame, axes=axes
    )
    pix = geom.get_idx()
    if axes is not None:
        idx = tuple([1] * len(axes))
        pix_img = geom.get_idx(idx=idx)
        m = np.all(np.stack([x == y for x, y in zip(idx, pix[2:])]), axis=0)
        m2 = pix_img[0] != -1
        assert_allclose(pix[0][m], np.ravel(pix_img[0][m2]))
        assert_allclose(pix[1][m], np.ravel(pix_img[1][m2]))


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skydir", "axes"), wcs_test_geoms
)
def test_wcsgeom_test_pix_to_coord(npix, binsz, frame, proj, skydir, axes):
    geom = WcsGeom.create(
        npix=npix, binsz=binsz, skydir=skydir, proj=proj, frame=frame, axes=axes
    )
    assert_allclose(geom.get_coord()[0], geom.pix_to_coord(geom.get_idx())[0])


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skydir", "axes"), wcs_test_geoms
)
def test_wcsgeom_test_coord_to_idx(npix, binsz, frame, proj, skydir, axes):
    geom = WcsGeom.create(npix=npix, binsz=binsz, proj=proj, frame=frame, axes=axes)
    assert_allclose(geom.get_idx()[0], geom.coord_to_idx(geom.get_coord())[0])

    if not geom.is_allsky:
        coords = geom.center_coord[:2] + tuple([ax.center[0] for ax in geom.axes])
        coords[0][...] += 2.0 * np.max(geom.width[0])
        idx = geom.coord_to_idx(coords)
        assert_allclose(np.full_like(coords[0].value, -1, dtype=int), idx[0])
        idx = geom.coord_to_idx(coords, clip=True)
        assert np.all(
            np.not_equal(np.full_like(coords[0].value, -1, dtype=int), idx[0])
        )


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skydir", "axes"), wcs_test_geoms
)
def test_wcsgeom_read_write(tmp_path, npix, binsz, frame, proj, skydir, axes):
    geom0 = WcsGeom.create(npix=npix, binsz=binsz, proj=proj, frame=frame, axes=axes)

    hdu_bands = geom0.to_bands_hdu(hdu_bands="TEST_BANDS")
    hdu_prim = fits.PrimaryHDU()
    hdu_prim.header.update(geom0.to_header())

    hdulist = fits.HDUList([hdu_prim, hdu_bands])
    hdulist.writeto(tmp_path / "tmp.fits")

    with fits.open(tmp_path / "tmp.fits", memmap=False) as hdulist:
        geom1 = WcsGeom.from_header(hdulist[0].header, hdulist["TEST_BANDS"])

    assert_allclose(geom0.npix, geom1.npix)
    assert geom0.frame == geom1.frame


@requires_data()
def test_wcsgeom_from_header():
    file_path = make_path("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
    with fits.open(file_path, memmap=False) as hdulist:
        geom2 = WcsGeom.from_header(hdulist[1].header, hdulist["COUNTS_BANDS"])

    assert geom2.wcs.naxis == 2
    assert_equal(geom2.wcs._naxis, (240, 320))


def test_wcsgeom_to_hdulist():
    npix, binsz, frame, proj, skydir, axes = wcs_test_geoms[3]
    geom = WcsGeom.create(npix=npix, binsz=binsz, proj=proj, frame=frame, axes=axes)

    hdu = geom.to_bands_hdu(hdu_bands="TEST")
    assert hdu.header["AXCOLS1"] == "E_MIN,E_MAX"
    assert hdu.header["AXCOLS2"] == "AXIS1_MIN,AXIS1_MAX"


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skydir", "axes"), wcs_test_geoms
)
def test_wcsgeom_contains(npix, binsz, frame, proj, skydir, axes):
    geom = WcsGeom.create(
        npix=npix, binsz=binsz, skydir=skydir, proj=proj, frame=frame, axes=axes
    )
    coords = geom.get_coord()
    m = np.isfinite(coords[0])
    coords = [c[m] for c in coords]
    assert_allclose(geom.contains(coords), np.ones(coords[0].shape, dtype=bool))

    if axes is not None:
        coords = [c[0] for c in coords[:2]] + [ax.edges[-1] + 1.0 for ax in axes]
        assert_allclose(geom.contains(coords), np.zeros((1,), dtype=bool))

    if not geom.is_allsky:
        coords = [0.0, 0.0] + [ax.center[0] for ax in geom.axes]
        assert_allclose(geom.contains(coords), np.zeros((1,), dtype=bool))


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skydir", "axes"), wcs_test_geoms
)
def test_wcs_geom_from_aligned(npix, binsz, frame, proj, skydir, axes):
    geom = WcsGeom.create(
        npix=npix, binsz=binsz, skydir=(0, 0), proj=proj, frame=frame, axes=axes
    )

    aligned_geom = WcsGeom.from_aligned(geom=geom, skydir=(2, 3), width="90 deg")

    assert aligned_geom.is_aligned(geom)


def test_from_aligned_vs_cutout():
    skydir = SkyCoord(0.12, -0.34, unit="deg", frame="galactic")

    geom = WcsGeom.create(binsz=0.1, skydir=skydir, proj="AIT", frame="galactic")

    position = SkyCoord("2.23d", "3.102d", frame="galactic")

    width = ("89 deg", "79 deg")
    aligned_geom = WcsGeom.from_aligned(geom=geom, skydir=position, width=width)

    geom_cutout = geom.cutout(position=position, width=width)

    assert geom_cutout == aligned_geom


def test_from_aligned_vs_cutout_tan():
    skydir = SkyCoord(0, 0, unit="deg", frame="galactic")

    geom = WcsGeom.create(
        binsz=1, skydir=skydir, proj="TAN", frame="galactic", width=("180d", "90d")
    )

    position = SkyCoord("53.23d", "22.102d", frame="galactic")
    width = ("17 deg", "15 deg")

    geom_cutout = geom.cutout(position=position, width=width, mode="partial")

    aligned_geom = WcsGeom.from_aligned(geom=geom, skydir=position, width=width)

    assert aligned_geom == geom_cutout


def test_wcsgeom_solid_angle():
    # Test using a CAR projection map with an extra axis
    binsz = 1.0 * u.deg
    npix = 10
    geom = WcsGeom.create(
        skydir=(0, 0),
        npix=(npix, npix),
        binsz=binsz,
        frame="galactic",
        proj="CAR",
        axes=[MapAxis.from_edges([0, 2, 3])],
    )

    solid_angle = geom.solid_angle()

    # Check array size
    assert solid_angle.shape == (1, npix, npix)

    # Test at b = 0 deg
    assert solid_angle.unit == "sr"
    assert_allclose(solid_angle.value[0, 5, 5], 0.0003046, rtol=1e-3)

    # Test at b = 5 deg
    assert_allclose(solid_angle.value[0, 9, 5], 0.0003038, rtol=1e-3)


def test_wcsgeom_solid_angle_symmetry():
    geom = WcsGeom.create(
        skydir=(0, 0), frame="galactic", npix=(3, 3), binsz=20.0 * u.deg
    )

    sa = geom.solid_angle()

    assert_allclose(sa[1, :], sa[1, 0])  # Constant along lon
    assert_allclose(sa[0, 1], sa[2, 1])  # Symmetric along lat
    with pytest.raises(AssertionError):
        # Not constant along lat due to changes in solid angle (great circle)
        assert_allclose(sa[:, 1], sa[0, 1])


def test_wcsgeom_solid_angle_ait():
    # Pixels that don't correspond to locations on the sky
    # should have solid angles set to NaN
    ait_geom = WcsGeom.create(
        skydir=(0, 0), width=(360, 180), binsz=20, frame="galactic", proj="AIT"
    )
    solid_angle = ait_geom.solid_angle().to_value("deg2")

    assert_allclose(solid_angle[4, 1], 397.04838)
    assert_allclose(solid_angle[4, 16], 397.751841)
    assert_allclose(solid_angle[1, 8], 381.556269)
    assert_allclose(solid_angle[7, 8], 398.34725)

    assert np.isnan(solid_angle[0, 0])


def test_wcsgeom_separation():
    geom = WcsGeom.create(
        skydir=(0, 0),
        npix=10,
        binsz=0.1,
        frame="galactic",
        proj="CAR",
        axes=[MapAxis.from_edges([0, 2, 3])],
    )
    position = SkyCoord(1, 0, unit="deg", frame="galactic").icrs
    separation = geom.separation(position)

    assert separation.unit == "deg"
    assert separation.shape == (10, 10)
    assert_allclose(separation.value[0, 0], 0.7106291438079875)

    # Make sure it also works for 2D maps as input
    separation = geom.to_image().separation(position)
    assert separation.unit == "deg"
    assert separation.shape == (10, 10)
    assert_allclose(separation.value[0, 0], 0.7106291438079875)


def test_cutout():
    geom = WcsGeom.create(
        skydir=(0, 0),
        npix=10,
        binsz=0.1,
        frame="galactic",
        proj="CAR",
        axes=[MapAxis.from_edges([0, 2, 3])],
    )
    position = SkyCoord(0.1, 0.2, unit="deg", frame="galactic")
    cutout_geom = geom.cutout(position=position, width=2 * 0.3 * u.deg, mode="trim")

    center_coord = cutout_geom.center_coord
    assert_allclose(center_coord[0].value, 0.1)
    assert_allclose(center_coord[1].value, 0.2)
    assert_allclose(center_coord[2].value, 2.0)

    assert cutout_geom.data_shape == (2, 6, 6)
    assert cutout_geom.data_shape_axes == (2, 1, 1)


def test_cutout_info():
    geom = WcsGeom.create(skydir=(0, 0), npix=10)
    position = SkyCoord(0, 0, unit="deg")
    cutout_geom = geom.cutout(position=position, width="2 deg")

    cutout_info = cutout_geom.cutout_slices(geom)

    assert cutout_info["parent-slices"][0].start == 3
    assert cutout_info["parent-slices"][1].start == 3

    assert cutout_info["cutout-slices"][0].start == 0
    assert cutout_info["cutout-slices"][1].start == 0


def test_cutout_min_size():
    geom = WcsGeom.create(skydir=(0, 0), npix=10, binsz=0.5)
    position = SkyCoord(0, 0, unit="deg")
    cutout_geom = geom.cutout(position=position, width=["2 deg", "0.1 deg"])
    assert cutout_geom.data_shape == (1, 4)

    cutout_geom = geom.cutout(position=position, width=["2 deg", "0.1 deg"], min_npix=3)
    assert cutout_geom.data_shape == (3, 4)

    geom = WcsGeom.create(skydir=(0, 0), npix=(10, 1), binsz=0.5)
    position = SkyCoord(0, 0, unit="deg")
    cutout_geom = geom.cutout(position=position, width=["1 deg", "0.1 deg"])
    assert cutout_geom.data_shape == (1, 2)

    cutout_geom = geom.cutout(position=position, width=["1 deg", "0.1 deg"], min_npix=3)
    assert cutout_geom.data_shape == (1, 3)


def test_wcs_geom_get_coord():
    geom = WcsGeom.create(
        skydir=(0, 0), npix=(4, 3), binsz=1, frame="galactic", proj="CAR"
    )
    coord = geom.get_coord(mode="edges")
    assert_allclose(coord.lon[0, 0].value, 2)
    assert coord.lon[0, 0].unit == "deg"
    assert_allclose(coord.lat[0, 0].value, -1.5)
    assert coord.lat[0, 0].unit == "deg"


def test_wcs_geom_instance_cache():
    geom_1 = WcsGeom.create(npix=(3, 3))
    geom_2 = WcsGeom.create(npix=(3, 3))

    coord_1, coord_2 = geom_1.get_coord(), geom_2.get_coord()

    coord_1_cached, coord_2_cached = geom_1.get_coord(), geom_2.get_coord()

    assert id(coord_1) == id(coord_1_cached)
    assert id(coord_2) == id(coord_2_cached)


def test_wcs_geom_squash():
    axis = MapAxis.from_nodes([1, 2, 3], name="test-axis")
    geom = WcsGeom.create(npix=(3, 3), axes=[axis])
    geom_squashed = geom.squash(axis_name="test-axis")
    assert geom_squashed.data_shape == (1, 3, 3)


def test_wcs_geom_drop():
    ax1 = MapAxis.from_nodes([1, 2, 3], name="ax1")
    ax2 = MapAxis.from_nodes([1, 2], name="ax2")
    ax3 = MapAxis.from_nodes([1, 2, 3, 4], name="ax3")
    geom = WcsGeom.create(npix=(3, 3), axes=[ax1, ax2, ax3])
    geom_drop = geom.drop(axis_name="ax1")
    assert geom_drop.data_shape == (4, 2, 3, 3)


def test_wcs_geom_resample_overflows():
    ax1 = MapAxis.from_edges([1, 2, 3, 4, 5], name="ax1")
    ax2 = MapAxis.from_nodes([1, 2, 3], name="ax2")
    geom = WcsGeom.create(npix=(3, 3), axes=[ax1, ax2])
    new_axis = MapAxis.from_edges([-1.0, 1, 2.3, 4.8, 6], name="ax1")
    geom_resample = geom.resample_axis(axis=new_axis)

    assert geom_resample.data_shape == (3, 2, 3, 3)
    assert geom_resample.data_shape_axes == (3, 2, 1, 1)
    assert_allclose(geom_resample.axes[0].edges, [1, 2, 5])


def test_wcs_geom_get_pix_coords():
    geom = WcsGeom.create(
        skydir=(0, 0), npix=(4, 3), binsz=1, frame="galactic", proj="CAR", axes=axes1
    )
    idx_center = geom.get_pix(mode="center")

    for idx in idx_center:
        assert idx.shape == (2, 3, 4)
        assert_allclose(idx[0, 0, 0], 0)

    idx_edge = geom.get_pix(mode="edges")
    for idx, desired in zip(idx_edge, [-0.5, -0.5, 0]):
        assert idx.shape == (2, 4, 5)
        assert_allclose(idx[0, 0, 0], desired)


def test_geom_str():
    geom = WcsGeom.create(
        skydir=(0, 0), npix=(10, 4), binsz=50, frame="galactic", proj="AIT"
    )

    str_info = str(geom)
    assert geom.__class__.__name__ in str_info
    assert "wcs ref" in str_info
    assert "center" in str_info
    assert "projection" in str_info
    assert "axes" in str_info
    assert "shape" in str_info
    assert "ndim" in str_info
    assert "width" in str_info


def test_geom_refpix():
    refpix = (400, 300)
    geom = WcsGeom.create(
        skydir=(0, 0), npix=(800, 600), refpix=refpix, binsz=0.1, frame="galactic"
    )
    assert_allclose(geom.wcs.wcs.crpix, refpix)


def test_region_mask():
    geom = WcsGeom.create(npix=(3, 3), binsz=2, proj="CAR")

    r1 = CircleSkyRegion(SkyCoord(0, 0, unit="deg"), 1 * u.deg)
    r2 = CircleSkyRegion(SkyCoord(20, 20, unit="deg"), 1 * u.deg)
    r3 = PointSkyRegion(SkyCoord(0.5, 0.5, unit="deg"))
    regions = [r1, r2]

    mask = geom.region_mask(regions)
    assert mask.data.dtype == bool
    assert np.sum(mask.data) == 1

    mask = geom.region_mask(regions, inside=False)
    assert np.sum(mask.data) == 8

    mask = geom.region_mask([r3])
    assert np.sum(mask.data) == 4
    assert mask.data.dtype == bool


def test_energy_mask():
    energy_axis = MapAxis.from_nodes(
        [1, 10, 100], interp="log", name="energy", unit="TeV"
    )
    geom = WcsGeom.create(npix=(1, 1), binsz=1, proj="CAR", axes=[energy_axis])

    mask = geom.energy_mask(energy_min=3 * u.TeV).data
    assert not mask[0, 0, 0]
    assert mask[1, 0, 0]
    assert mask[2, 0, 0]

    mask = geom.energy_mask(energy_max=30 * u.TeV).data
    assert mask[0, 0, 0]
    assert not mask[1, 0, 0]
    assert not mask[2, 0, 0]

    mask = geom.energy_mask(energy_min=3 * u.TeV, energy_max=40 * u.TeV).data
    assert not mask[0, 0, 0]
    assert not mask[2, 0, 0]
    assert mask[1, 0, 0]


def test_boundary_mask():
    axis = MapAxis.from_edges([1, 10, 100])
    geom = WcsGeom.create(
        skydir=(0, 0),
        binsz=0.02,
        width=(2, 2),
        axes=[axis],
    )

    mask = geom.boundary_mask(width=(0.3 * u.deg, 0.1 * u.deg))
    assert np.sum(mask.data[0, :, :]) == 6300
    assert np.sum(mask.data[1, :, :]) == 6300


@pytest.mark.parametrize(
    ("width", "out"),
    [
        (10, (10, 10)),
        ((10 * u.deg).to("rad"), (10, 10)),
        ((10, 5), (10, 5)),
        (("10 deg", "5 deg"), (10, 5)),
        (Angle([10, 5], "deg"), (10, 5)),
        ((10 * u.deg, 5 * u.deg), (10, 5)),
        ((10, 5) * u.deg, (10, 5)),
        ([10, 5], (10, 5)),
        (["10 deg", "5 deg"], (10, 5)),
        (np.array([10, 5]), (10, 5)),
    ],
)
def test_check_width(width, out):
    width = _check_width(width)
    assert isinstance(width, tuple)
    assert isinstance(width[0], float)
    assert isinstance(width[1], float)
    assert width == out

    geom = WcsGeom.create(width=width, binsz=1.0)
    assert tuple(geom.npix) == out


def test_check_binsz():
    # float
    binsz = _check_binsz(0.1)
    assert isinstance(binsz, float)
    # string and other units
    binsz = _check_binsz("0.1deg")
    assert isinstance(binsz, float)
    binsz = _check_binsz("3.141592653589793 rad")
    assert_allclose(binsz, 180)
    # tuple
    binsz = _check_binsz(("0.1deg", "0.2deg"))
    assert isinstance(binsz, tuple)
    assert isinstance(binsz[0], float)
    assert isinstance(binsz[1], float)
    # list
    binsz = _check_binsz(["0.1deg", "0.2deg"])
    assert isinstance(binsz, list)
    assert isinstance(binsz[0], float)
    assert isinstance(binsz[1], float)


def test_check_width_bad_input():
    with pytest.raises(IndexError):
        _check_width(width=(10,))


def test_get_axis_index_by_name():
    e_axis = MapAxis.from_edges([1, 5], name="energy")
    geom = WcsGeom.create(width=5, binsz=1.0, axes=[e_axis])
    assert geom.axes.index("energy") == 0
    with pytest.raises(ValueError):
        geom.axes.index("time")


test_axis1 = [MapAxis(nodes=(1, 2, 3, 4), unit="TeV", node_type="center")]
test_axis2 = [
    MapAxis(nodes=(1, 2, 3, 4), unit="TeV", node_type="center"),
    MapAxis(nodes=(1, 2, 3), unit="TeV", node_type="center"),
]

skydir2 = SkyCoord(110.0, 75.0 + 1e-8, unit="deg", frame="icrs")
skydir3 = SkyCoord(110.0, 75.0 + 1e-3, unit="deg", frame="icrs")

compatibility_test_geoms = [
    (10, 0.1, "galactic", "CAR", skydir, test_axis1, True),
    (10, 0.1, "galactic", "CAR", skydir2, test_axis1, True),
    (10, 0.1, "galactic", "CAR", skydir3, test_axis1, False),
    (10, 0.1, "galactic", "TAN", skydir, test_axis1, False),
    (8, 0.1, "galactic", "CAR", skydir, test_axis1, False),
    (10, 0.1, "galactic", "CAR", skydir, test_axis2, False),
    (10, 0.1, "galactic", "CAR", skydir.galactic, test_axis1, True),
]


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skypos", "axes", "result"),
    compatibility_test_geoms,
)
def test_wcs_geom_equal(npix, binsz, frame, proj, skypos, axes, result):
    geom0 = WcsGeom.create(
        skydir=skydir, npix=10, binsz=0.1, proj="CAR", frame="galactic", axes=test_axis1
    )
    geom1 = WcsGeom.create(
        skydir=skypos, npix=npix, binsz=binsz, proj=proj, frame=frame, axes=axes
    )

    assert (geom0 == geom1) is result
    assert (geom0 != geom1) is not result


def test_irregular_geom_equality():
    axis = MapAxis.from_bounds(1, 3, 10, name="axis", unit="")
    geom0 = WcsGeom.create(skydir=(0, 0), npix=10, binsz=0.1, axes=[axis])
    binsizes = np.ones((10)) * 0.1
    geom1 = WcsGeom.create(skydir=(0, 0), npix=10, binsz=binsizes, axes=[axis])

    with pytest.raises(NotImplementedError):
        geom0 == geom1


def test_wcs_geom_pad():
    axis = MapAxis.from_bounds(0, 1, nbin=1, name="axis", unit="")
    geom = WcsGeom.create(skydir=(0, 0), npix=10, binsz=0.1, axes=[axis])

    geom_pad = geom.pad(axis_name="axis", pad_width=1)
    assert_allclose(geom_pad.axes["axis"].edges, [-1, 0, 1, 2])


@pytest.mark.parametrize("node_type", ["edges", "center"])
@pytest.mark.parametrize("interp", ["log", "lin", "sqrt"])
def test_read_write(tmp_path, node_type, interp):
    # Regression test for MapAxis interp and node_type FITS serialization
    # https://github.com/gammapy/gammapy/issues/1887
    e_ax = MapAxis([1, 2], interp, "energy", node_type, "TeV")
    t_ax = MapAxis([3, 4], interp, "time", node_type, "s")
    m = Map.create(binsz=1, npix=10, axes=[e_ax, t_ax], unit="m2")

    # Check what Gammapy writes in the FITS header
    header = m.to_hdu().header
    assert header["INTERP1"] == interp
    assert header["INTERP2"] == interp

    # Check that all MapAxis properties are preserved on FITS I/O
    m.write(tmp_path / "tmp.fits", overwrite=True)
    m2 = Map.read(tmp_path / "tmp.fits")
    assert m2.geom == m.geom


@pytest.mark.parametrize(
    ("npix", "binsz", "frame", "proj", "skypos", "axes", "result"),
    compatibility_test_geoms,
)
def test_wcs_geom_to_binsz(npix, binsz, frame, proj, skypos, axes, result):
    geom = WcsGeom.create(
        skydir=skydir, npix=10, binsz=0.1, proj="CAR", frame="galactic", axes=test_axis1
    )

    geom_new = geom.to_binsz(binsz=0.5)

    assert_allclose(geom_new.pixel_scales.value, 0.5)


def test_non_equal_binsz():
    geom = WcsGeom.create(
        width=(360, 180), binsz=(360, 60), frame="icrs", skydir=(0, 0), proj="CAR"
    )

    coords = geom.get_coord()

    assert_allclose(coords["lon"].to_value("deg"), 0)
    assert_allclose(coords["lat"].to_value("deg"), [[-60], [0], [60]])


def test_wcs_geom_non_equal_dim():
    geom = WcsGeom.create(skydir=(0, 0), binsz=1, width=(3, 5))

    assert geom.data_shape == (5, 3)
    assert geom.data_shape == geom.wcs.array_shape


def test_wcs_geom_to_even_npix():
    geom = WcsGeom.create(skydir=(0, 0), binsz=1, width=(3, 3))

    geom_even = geom.to_even_npix()

    assert geom_even.data_shape == (4, 4)


def test_wcs_geom_no_zero_shape_cut():
    geom = WcsGeom.create(skydir=(0, 2.5), binsz=(360, 5), width=(360, 180))
    skydir = SkyCoord(110.0, 75.0, unit="deg", frame="icrs")

    geom_cut = geom.cutout(skydir, width=5)

    assert geom_cut.data_shape != (1, 0)
    assert geom_cut.data_shape == (1, 1)


def test_wcs_geom_with_timeaxis():
    center = SkyCoord("0d", "0d", frame="icrs")

    t_ref = Time(55555, format="mjd")

    energy_axis = MapAxis.from_energy_bounds(
        "1 TeV", "10 TeV", nbin=3, name="energy_true"
    )

    time_min = t_ref + [1, 3, 5, 7] * u.day
    time_max = t_ref + [2, 4, 6, 8] * u.day

    time_axis = TimeMapAxis.from_time_edges(time_min=time_min, time_max=time_max)

    wcs_geom = WcsGeom.create(
        width=[1, 1.2], binsz=0.2, skydir=center, axes=[energy_axis, time_axis]
    )

    coords = wcs_geom.get_coord()
    assert coords.shape == (4, 3, 6, 5)
    assert_allclose(
        coords[0][0][0][0:-1:5].value,
        [[4.000e-01, 2.000e-01, 0.000e00, 3.598e02, 3.596e02]],
        rtol=1e-5,
    )


@requires_dependency("ray")
def test_instance_cache():
    cache.USE_INSTANCE_CACHE = True

    kwargs = dict(skydir=(83.63, 22.01), width=5, binsz=0.02)

    geom1 = WcsGeom.create(**kwargs)
    geom2 = WcsGeom.create(**kwargs)
    geom3 = WcsGeom.create(width=5, binsz=0.02, skydir=(83.63, 22.01))  # reordered

    assert geom1 is geom2
    assert geom1 is geom3
    assert len(list(WcsGeom._instances._dict[id(WcsGeom)][0].keys())) == 1

    geom4 = WcsGeom.create(width=5, binsz=0.05, skydir=(83.63, 22.01))  # different
    assert geom4 is not geom1
    assert len(list(WcsGeom._instances._dict[id(WcsGeom)][0].keys())) == 2

    cache.USE_INSTANCE_CACHE = False

    kwargs = dict(skydir=(83.63, 22.01), width=5, binsz=0.02)

    geom1 = WcsGeom.create(**kwargs)
    geom2 = WcsGeom.create(**kwargs)

    assert geom1 is not geom2


def test_caplog(caplog):
    geom1 = WcsGeom.create(width=5, binsz=0.05, skydir=(83.63, 22.01))
    geom2 = WcsGeom.create(width=1, binsz=0.05, skydir=(83.63, 22.01))

    caplog.set_level(logging.DEBUG)
    assert geom1 != geom2
    assert "WcsGeom data shape is not equal" in [_.message for _ in caplog.records]
    assert "DEBUG" in [_.levelname for _ in caplog.records]

    geom2 = WcsGeom.create(width=5, binsz=0.05, skydir=(8.63, 22.01))
    assert geom1 != geom2
    assert "WcsGeom wcs is not equal" in [_.message for _ in caplog.records]
    assert "DEBUG" in [_.levelname for _ in caplog.records]

    axis1 = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=3, name="energy_true")
    axis2 = MapAxis.from_energy_bounds("2 TeV", "10 TeV", nbin=3, name="energy")
    geom1 = WcsGeom.create(width=5, binsz=0.05, skydir=(83.63, 22.01), axes=[axis1])
    geom2 = WcsGeom.create(width=5, binsz=0.05, skydir=(83.63, 22.01), axes=[axis2])
    assert geom1 != geom2
    assert "WcsGeom axes are not equal" in [_.message for _ in caplog.records]
    assert "DEBUG" in [_.levelname for _ in caplog.records]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os
import pytest
import numpy as np
from numpy.testing import assert_allclose, assert_equal
from astropy.coordinates import SkyCoord
from gammapy.datasets import Datasets, SpectrumDatasetOnOff
from gammapy.datasets.tests.test_map import get_map_dataset
from gammapy.maps import MapAxis, WcsGeom
from gammapy.modeling import Fit
from gammapy.modeling.models import FoVBackgroundModel, Models, SkyModel
from gammapy.modeling.tests.test_fit import MyDataset
from gammapy.utils.testing import requires_data


@pytest.fixture(scope="session")
def datasets():
    return Datasets([MyDataset(name="test-1"), MyDataset(name="test-2")])


def test_datasets_init(datasets):
    # Passing a Python list of `Dataset` objects should work
    Datasets(list(datasets))

    # Passing an existing `Datasets` object should work
    Datasets(datasets)


def test_datasets_types(datasets):
    assert datasets.is_all_same_type


def test_datasets_likelihood(datasets):
    likelihood = datasets.stat_sum()
    assert_allclose(likelihood, 14472200.0002)


def test_datasets_str(datasets):
    assert "Datasets" in str(datasets)


def test_datasets_getitem(datasets):
    assert datasets["test-1"].name == "test-1"
    assert datasets["test-2"].name == "test-2"


def test_names(datasets):
    assert datasets.names == ["test-1", "test-2"]


def test_datasets_mutation():
    dat = MyDataset(name="test-1")
    dats = Datasets([MyDataset(name="test-2"), MyDataset(name="test-3")])
    dats2 = Datasets([MyDataset(name="test-4"), MyDataset(name="test-5")])

    dats.insert(0, dat)
    assert dats.names == ["test-1", "test-2", "test-3"]

    dats.extend(dats2)
    assert dats.names == ["test-1", "test-2", "test-3", "test-4", "test-5"]

    dat3 = dats[3]
    dats.remove(dats[3])
    assert dats.names == ["test-1", "test-2", "test-3", "test-5"]
    dats.append(dat3)
    assert dats.names == ["test-1", "test-2", "test-3", "test-5", "test-4"]
    dats.pop(3)
    assert dats.names == ["test-1", "test-2", "test-3", "test-4"]

    with pytest.raises(ValueError, match="Dataset names must be unique"):
        dats.append(dat)

    with pytest.raises(ValueError, match="Dataset names must be unique"):
        dats.insert(0, dat)

    with pytest.raises(ValueError, match="Dataset names must be unique"):
        dats.extend(dats2)


@requires_data()
def test_datasets_info_table():
    datasets_hess = Datasets()

    for obs_id in [23523, 23526]:
        dataset = SpectrumDatasetOnOff.read(
            f"$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs{obs_id}.fits"
        )
        datasets_hess.append(dataset)

    table = datasets_hess.info_table()
    assert table["name"][0] == "23523"
    assert table["name"][1] == "23526"
    assert np.isnan(table["npred_signal"][0])
    assert_equal(table["counts"], [124, 126])
    assert_allclose(table["background"], [7.666, 8.583], rtol=1e-3)

    table_cumul = datasets_hess.info_table(cumulative=True)
    assert table_cumul["name"][0] == "stacked"
    assert table_cumul["name"][1] == "stacked"
    assert np.isnan(table_cumul["npred_signal"][0])
    assert table_cumul["alpha"][1] == table_cumul["alpha"][0]
    assert_equal(table_cumul["counts"], [124, 250])
    assert_allclose(table_cumul["background"], [7.666, 16.25], rtol=1e-3)

    assert table["excess"].sum() == table_cumul["excess"][1]
    assert table["counts"].sum() == table_cumul["counts"][1]
    assert table["background"].sum() == table_cumul["background"][1]

    datasets_hess[0].mask_safe.data = ~datasets_hess[0].mask_safe.data
    assert datasets_hess.info_table()["counts"][0] == 65

    datasets_hess[0].mask_safe.data = np.ones_like(
        datasets_hess[0].mask_safe.data, dtype=bool
    )
    datasets_hess[1].mask_safe.data = np.ones_like(
        datasets_hess[1].mask_safe.data, dtype=bool
    )
    assert_equal(datasets_hess.info_table()["counts"], [189, 199])
    assert_equal(datasets_hess.info_table(cumulative=True)["counts"], [189, 388])


@requires_data()
def test_datasets_write(tmp_path):
    axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=2)
    geom = WcsGeom.create(
        skydir=(266.40498829, -28.93617776),
        binsz=0.02,
        width=(2, 2),
        frame="icrs",
        axes=[axis],
    )

    axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=3, name="energy_true")
    geom_etrue = WcsGeom.create(
        skydir=(266.40498829, -28.93617776),
        binsz=0.02,
        width=(2, 2),
        frame="icrs",
        axes=[axis],
    )

    dataset_1 = get_map_dataset(geom, geom_etrue, name="test-1")
    datasets = Datasets([dataset_1])

    model = SkyModel.create("pl", "point", name="src")
    model.spatial_model.position = SkyCoord("266d", "-28.93d", frame="icrs")

    dataset_1.models = [model]

    datasets.write(
        filename=tmp_path / "test",
        filename_models=tmp_path / "test_model",
        overwrite=False,
    )
    os.remove(tmp_path / "test-1.fits")

    with pytest.raises(OSError):
        datasets.write(
            filename=tmp_path / "test",
            filename_models=tmp_path / "test_model",
            overwrite=False,
        )


@requires_data()
def test_datasets_fit():
    axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=2)
    geom = WcsGeom.create(
        skydir=(266.40498829, -28.93617776),
        binsz=0.05,
        width=(20, 20),
        frame="icrs",
        axes=[axis],
    )

    axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=3, name="energy_true")
    geom_etrue = WcsGeom.create(
        skydir=(266.40498829, -28.93617776),
        binsz=0.05,
        width=(20, 20),
        frame="icrs",
        axes=[axis],
    )

    dataset_1 = get_map_dataset(geom, geom_etrue, name="test-1")
    dataset_1.mask_fit = None
    dataset_1.background /= 400
    dataset_2 = get_map_dataset(geom, geom_etrue, name="test-2")
    dataset_2.mask_fit = None
    dataset_2.background /= 400

    datasets = Datasets([dataset_1, dataset_2])

    model = SkyModel.create("pl", "point", name="src")
    model.spatial_model.position = dataset_1.exposure.geom.center_skydir

    model2 = model.copy()
    model2.spatial_model.lon_0.value += 0.1
    model2.spatial_model.lat_0.value += 0.1

    models = Models(
        [
            model,
            model2,
            FoVBackgroundModel(dataset_name=dataset_1.name),
            FoVBackgroundModel(dataset_name=dataset_2.name),
        ]
    )

    datasets.models = models
    dataset_1.fake()
    dataset_2.fake()

    fit = Fit()
    results = fit.run(datasets)

    assert_allclose(results.models.covariance.data, datasets.models.covariance.data)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from typing import ClassVar, Literal, Optional
import pytest
from numpy.testing import assert_allclose
from astropy.coordinates import AltAz, SkyCoord
from astropy.io import fits
from astropy.time import Time
from pydantic import ValidationError
from gammapy.utils.metadata import (
    METADATA_FITS_KEYS,
    CreatorMetaData,
    MetaData,
    ObsInfoMetaData,
    PointingInfoMetaData,
    TargetMetaData,
    TimeInfoMetaData,
)
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data


@pytest.fixture()
def hess_eventlist_header():
    filename = make_path(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_023523.fits.gz"
    )
    hdulist = fits.open(filename)
    return hdulist["EVENTS"].header


def test_creator():
    default = CreatorMetaData(date="2022-01-01", creator="gammapy", origin="CTA")

    assert default.creator == "gammapy"
    assert default.origin == "CTA"
    assert_allclose(default.date.mjd, 59580)

    default.creator = "other"
    assert default.creator == "other"

    with pytest.raises(ValidationError):
        default.date = 3


def test_creator_to_header():
    header = CreatorMetaData(
        date="2022-01-01", creator="gammapy", origin="CTA"
    ).to_header(format="gadf")

    assert header["CREATOR"] == "gammapy"
    assert header["ORIGIN"] == "CTA"
    assert header["CREATED"] == "2022-01-01 00:00:00.000"


def test_creator_from_incorrect_header():
    # Create header with a 'bad' date
    hdu = fits.PrimaryHDU()
    hdu.header["CREATOR"] = "gammapy"
    hdu.header["CREATED"] = "Tues 6 Feb"

    meta = CreatorMetaData.from_header(hdu.header)

    assert meta.date == hdu.header["CREATED"]
    assert meta.creator == hdu.header["CREATOR"]


def test_subclass():
    class TestMetaData(MetaData):
        _tag: ClassVar[Literal["tag"]] = "tag"
        name: str
        mode: Optional[SkyCoord] = None
        creation: Optional[CreatorMetaData] = None

    creator = CreatorMetaData()
    test_meta = TestMetaData(name="test", creation=creator)

    assert test_meta.tag == "tag"

    assert test_meta.name == "test"
    assert test_meta.creation.creator.split()[0] == "Gammapy"

    with pytest.raises(ValidationError):
        test_meta.mode = "coord"

    yaml_str = test_meta.to_yaml()
    assert "name: test" in yaml_str
    assert "creation:" in yaml_str


def test_obs_info():
    obs_info = ObsInfoMetaData(obs_id="23523")

    assert obs_info.telescope is None
    assert obs_info.obs_id == 23523

    obs_info.obs_id = 23523
    assert obs_info.obs_id == 23523

    with pytest.raises(ValidationError):
        obs_info.obs_id = "ab"

    obs_info.instrument = "CTA-North"
    assert obs_info.instrument == "CTA-North"


@requires_data()
def test_obs_info_from_header(hess_eventlist_header):
    meta = ObsInfoMetaData.from_header(hess_eventlist_header, format="gadf")

    assert meta.telescope == "HESS"
    assert meta.obs_id == 23523
    assert meta.observation_mode == "WOBBLE"
    assert meta.sub_array is None


def test_obs_info_to_header():
    obs_info = ObsInfoMetaData(obs_id=23523, telescope="CTA-South")

    header = obs_info.to_header("gadf")

    assert header["OBS_ID"] == 23523
    assert header["TELESCOP"] == "CTA-South"
    assert "OBS_MODE" not in header


def test_pointing_info():
    position = SkyCoord(83.6287, 22.0147, unit="deg", frame="icrs")
    altaz = AltAz("20 deg", "45 deg")

    pointing = PointingInfoMetaData(radec_mean=position, altaz_mean=altaz)

    assert isinstance(pointing.altaz_mean, SkyCoord)
    assert_allclose(pointing.altaz_mean.alt.deg, 45.0)

    assert_allclose(pointing.radec_mean.ra.deg, 83.6287)

    pointing.radec_mean = position.galactic
    assert_allclose(pointing.radec_mean.ra.deg, 83.6287)

    with pytest.raises(ValidationError):
        pointing.radec_mean = altaz


def test_pointing_info_to_header():
    position = SkyCoord(83.6287, 22.0147, unit="deg", frame="icrs")
    altaz = AltAz("20 deg", "45 deg")

    header = PointingInfoMetaData(radec_mean=position, altaz_mean=altaz).to_header(
        "gadf"
    )

    assert_allclose(header["RA_PNT"], 83.6287)
    assert_allclose(header["AZ_PNT"], 20.0)

    header = PointingInfoMetaData(radec_mean=position).to_header("gadf")
    assert "AZ_PNT" not in header.keys()

    with pytest.raises(ValueError):
        PointingInfoMetaData(radec_mean=position, altaz_mean=altaz).to_header("bad")


@requires_data()
def test_pointing_info_from_header(hess_eventlist_header):
    meta = PointingInfoMetaData.from_header(hess_eventlist_header, format="gadf")

    assert_allclose(meta.radec_mean.ra.deg, 83.633333)
    assert_allclose(meta.altaz_mean.alt.deg, 41.389789)

    meta = PointingInfoMetaData.from_header({})
    assert meta.altaz_mean is None
    assert meta.radec_mean is None


def test_target_metadata():
    meta = TargetMetaData(
        name="center", position=SkyCoord(0.0, 0.0, unit="deg", frame="galactic")
    )
    header = meta.to_header(format="gadf")

    assert meta.name == "center"
    assert_allclose(meta.position.ra.deg, 266.404988)

    assert header["OBJECT"] == "center"
    assert_allclose(header["RA_OBJ"], 266.404988)

    header = TargetMetaData(name="center").to_header("gadf")
    assert header["OBJECT"] == "center"
    assert "RA_OBJ" not in header.keys()


@requires_data()
def test_target_metadata_from_header(hess_eventlist_header):
    meta = TargetMetaData.from_header(hess_eventlist_header, format="gadf")

    assert meta.name == "Crab Nebula"
    assert_allclose(meta.position.ra.deg, 83.63333333)


def test_time_info_metadata():
    meta = TimeInfoMetaData(
        reference_time="2023-01-01 00:00:00",
        time_start="2024-01-01 00:00:00",
        time_stop="2024-01-01 12:00:00",
    )

    assert isinstance(meta.reference_time, Time)
    delta = meta.time_stop - meta.time_start
    assert_allclose(delta.to_value("h"), 12)

    header = meta.to_header(format="gadf")
    assert header["MJDREFI"] == 59945
    assert header["TIMESYS"] == "tt"
    assert_allclose(header["TSTART"], 31536000.000000257)


@requires_data()
def test_time_info_metadata_from_header(hess_eventlist_header):
    meta = TimeInfoMetaData.from_header(hess_eventlist_header, format="gadf")

    assert_allclose(meta.reference_time.mjd, 51910.00074287037)
    assert_allclose(meta.time_start.mjd, 53343.92234009259)


def test_subclass_to_from_header():
    class TestMetaData(MetaData):
        _tag: ClassVar[Literal["test"]] = "test"
        creation: Optional[CreatorMetaData]
        pointing: Optional[PointingInfoMetaData]

    METADATA_FITS_KEYS.update({"test": {}})

    creator = CreatorMetaData(date="2022-01-01", creator="gammapy", origin="CTA")
    position = SkyCoord(83.6287, 22.0147, unit="deg", frame="icrs")
    altaz = AltAz("20 deg", "45 deg")

    pointing = PointingInfoMetaData(radec_mean=position, altaz_mean=altaz)

    test_meta = TestMetaData(pointing=pointing, creation=creator)

    header = test_meta.to_header()

    assert header["CREATOR"] == "gammapy"
    assert header["ORIGIN"] == "CTA"
    assert_allclose(header["RA_PNT"], 83.6287)
    assert_allclose(header["AZ_PNT"], 20.0)

    new = TestMetaData.from_header(header)

    assert new.creation.creator == "gammapy"
    assert_allclose(new.creation.date.decimalyear, 2022)
    assert_allclose(new.pointing.radec_mean.ra.deg, 83.6287)
    assert_allclose(new.pointing.altaz_mean.alt.deg, 45)
    # no new attributes allowed
    with pytest.raises(ValidationError):
        test_meta.extra = 3



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from pathlib import Path
import pytest
from astropy.coordinates import Angle
from astropy.time import Time
from astropy.units import Quantity
from pydantic import ValidationError
from gammapy.utils.testing import assert_allclose
from gammapy.workflow.config import GeneralConfig, WorkflowConfig

CONFIG_PATH = Path(__file__).resolve().parent.parent / "config"
DOC_FILE = CONFIG_PATH / "docs.yaml"


def test_config_default_types():
    config = WorkflowConfig()
    assert config.observations.obs_cone.frame is None
    assert config.observations.obs_cone.lon is None
    assert config.observations.obs_cone.lat is None
    assert config.observations.obs_cone.radius is None
    assert config.observations.obs_time.start is None
    assert config.observations.obs_time.stop is None
    assert config.datasets.geom.wcs.skydir.frame is None
    assert config.datasets.geom.wcs.skydir.lon is None
    assert config.datasets.geom.wcs.skydir.lat is None
    assert isinstance(config.datasets.geom.wcs.binsize, Angle)
    assert isinstance(config.datasets.geom.wcs.binsize_irf, Angle)
    assert isinstance(config.datasets.geom.axes.energy.min, Quantity)
    assert isinstance(config.datasets.geom.axes.energy.max, Quantity)
    assert isinstance(config.datasets.geom.axes.energy_true.min, Quantity)
    assert isinstance(config.datasets.geom.axes.energy_true.max, Quantity)
    assert isinstance(config.datasets.geom.selection.offset_max, Angle)
    assert config.fit.fit_range.min is None
    assert config.fit.fit_range.max is None
    assert isinstance(config.excess_map.correlation_radius, Angle)
    assert config.excess_map.energy_edges.min is None
    assert config.excess_map.energy_edges.max is None
    assert config.excess_map.energy_edges.nbins is None


def test_config_not_default_types():
    config = WorkflowConfig()
    config.observations.obs_cone = {
        "frame": "galactic",
        "lon": "83.633 deg",
        "lat": "22.014 deg",
        "radius": "1 deg",
    }
    config.fit.fit_range = {"min": "0.1 TeV", "max": "10 TeV"}
    assert config.observations.obs_cone.frame == "galactic"
    assert isinstance(config.observations.obs_cone.lon, Angle)
    assert isinstance(config.observations.obs_cone.lat, Angle)
    assert isinstance(config.observations.obs_cone.radius, Angle)
    config.observations.obs_time.start = "2019-12-01"
    assert isinstance(config.observations.obs_time.start, Time)
    with pytest.raises(ValueError):
        config.flux_points.energy.min = "1 deg"
    assert isinstance(config.fit.fit_range.min, Quantity)
    assert isinstance(config.fit.fit_range.max, Quantity)


def test_config_basics():
    config = WorkflowConfig()
    assert "WorkflowConfig" in str(config)
    config = WorkflowConfig.read(DOC_FILE)
    assert config.general.outdir == "."


def test_config_create_from_dict():
    data = {"general": {"log": {"level": "warning"}}}
    config = WorkflowConfig(**data)
    assert config.general.log.level == "warning"


def test_config_create_from_yaml():
    config = WorkflowConfig.read(DOC_FILE)
    assert isinstance(config.general, GeneralConfig)
    config_str = Path(DOC_FILE).read_text()
    config = WorkflowConfig.from_yaml(config_str)
    assert isinstance(config.general, GeneralConfig)


def test_config_to_yaml(tmp_path):
    config = WorkflowConfig()
    assert "level: info" in config.to_yaml()
    config = WorkflowConfig()
    fpath = Path(tmp_path) / "temp.yaml"
    config.write(fpath)
    text = Path(fpath).read_text()
    assert "stack" in text
    with pytest.raises(IOError):
        config.write(fpath)


def test_get_doc_sections():
    config = WorkflowConfig()
    doc = config._get_doc_sections()
    assert "general" in doc.keys()


def test_safe_mask_config_validation():
    config = WorkflowConfig()
    # Check empty list is accepted
    config.datasets.safe_mask.methods = []

    with pytest.raises(ValidationError):
        config.datasets.safe_mask.methods = ["bad"]


def test_time_range_iso():
    cfg = """
    observations:
        datastore: $GAMMAPY_DATA/hess-dl3-dr1
        obs_ids: [23523, 23526]
        obs_time: {
            start: [2004-12-04 22:04:48.000, 2004-12-04 22:26:24.000, 2004-12-04 22:53:45.600],
            stop: [2004-12-04 22:26:24.000, 2004-12-04 22:53:45.600, 2004-12-04 23:31:12.000]
            }
    """
    config = WorkflowConfig.from_yaml(cfg)

    assert_allclose(
        config.observations.obs_time.start.mjd, [53343.92, 53343.935, 53343.954]
    )


def test_time_range_jyear():
    cfg = """
    observations:
        datastore: $GAMMAPY_DATA/hess-dl3-dr1
        obs_ids: [23523, 23526]
        obs_time: {
            start: [J2004.92654346, J2004.92658453, J2004.92663655],
            stop: [J2004.92658453, J2004.92663655, J2004.92670773]
            }
    """
    config = WorkflowConfig.from_yaml(cfg)

    assert_allclose(
        config.observations.obs_time.start.mjd, [53343.92, 53343.935, 53343.954]
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import astropy.units as u
import numpy as np
from astropy.coordinates import (
    AltAz,
    Angle,
    BaseCoordinateFrame,
    CoordinateAttribute,
    DynamicMatrixTransform,
    EarthLocationAttribute,
    FunctionTransform,
    ICRS,
    RepresentationMapping,
    SkyCoord,
    SkyOffsetFrame,
    TimeAttribute,
    UnitSphericalRepresentation,
    frame_transform_graph,
)
from astropy.coordinates.matrix_utilities import matrix_transpose, rotation_matrix

__all__ = ["FoVAltAzFrame", "FoVICRSFrame", "fov_to_sky", "sky_to_fov"]

reflect_lon_matrix = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 1]])


class FoVAltAzFrame(BaseCoordinateFrame):
    """
    FoV coordinate frame. Centered on `origin` and aligned on AltAz frame at `location` and `obstime`.

    Longitudes are reversed.

    Attributes
    ----------
    origin: `~astropy.coordinates.AltAz`
        Origin of this frame as an Altaz coordinate
    obstime: `~astropy.time.Time`
        Observation time
    location: `~astropy.coordinates.EarthLocation`
        Location of the telescope/instrument/observatory

    Examples
    --------
    .. testcode::

        import astropy.units as u
        from astropy.time import Time
        from astropy.coordinates import AltAz, SkyCoord
        from gammapy.utils.observers import observatory_locations
        from gammapy.utils.coordinates import FoVAltAzFrame

        location = observatory_locations.get("ctao_north")
        obstime = Time("2025-01-01T00:00:00")
        origin = AltAz(alt=80*u.deg, az=172*u.deg, location=location, obstime=obstime)

        fov_frame = FoVAltAzFrame(origin=origin, location=location, obstime=obstime)

        crab= SkyCoord(83.63333333, 22.01444444, unit="deg", frame="icrs")
        crab.transform_to(fov_frame)

    """

    frame_specific_representation_info = {
        UnitSphericalRepresentation: [
            RepresentationMapping("lon", "fov_lon"),
            RepresentationMapping("lat", "fov_lat"),
        ]
    }
    default_representation = UnitSphericalRepresentation

    origin = CoordinateAttribute(default=None, frame=AltAz)

    obstime = TimeAttribute(default=None)
    location = EarthLocationAttribute(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # make sure telescope coordinate is in range [-180°, 180°]
        if isinstance(self._data, UnitSphericalRepresentation):
            self._data.lon.wrap_angle = Angle(180, unit=u.deg)


@frame_transform_graph.transform(FunctionTransform, FoVAltAzFrame, FoVAltAzFrame)
def fov_to_fov(from_fov_altaz_coord, to_fov_altaz_frame):
    """Transform between two `FoVFrame`."""
    intermediate_from = from_fov_altaz_coord.transform_to(from_fov_altaz_coord.origin)
    intermediate_to = intermediate_from.transform_to(to_fov_altaz_frame.origin)
    return intermediate_to.transform_to(to_fov_altaz_frame)


@frame_transform_graph.transform(DynamicMatrixTransform, AltAz, FoVAltAzFrame)
def altaz_to_fov_altaz(altaz_coord, fov_altaz_frame):
    """Convert a reference coordinate to a sky offset frame."""
    # Define rotation matrices along the position angle vector, and
    # relative to the origin.
    origin = fov_altaz_frame.origin.represent_as(UnitSphericalRepresentation)
    mat1 = rotation_matrix(-origin.lat, "y")
    mat2 = rotation_matrix(origin.lon, "z")

    return reflect_lon_matrix @ mat1 @ mat2


@frame_transform_graph.transform(DynamicMatrixTransform, FoVAltAzFrame, AltAz)
def fov_to_altaz(fov_altaz_coord, altaz_frame):
    """Convert an sky offset frame coordinate to the reference frame."""
    # use the forward transform, but just invert it
    mat = altaz_to_fov_altaz(altaz_frame, fov_altaz_coord)
    return matrix_transpose(mat)


class FoVICRSFrame(BaseCoordinateFrame):
    """
    FoV coordinate frame aligned on ICRS frame. Centered on `origin` an ICRS coordinate and aligned on ICRS frame.

    Longitudes are reversed.

    Attributes
    ----------
    origin: `~astropy.coordinates.ICRS`
        Origin of this frame as an ICRS coordinate

    Examples
    --------
    .. testcode::

        import astropy.units as u
        from astropy.time import Time
        from astropy.coordinates import SkyCoord
        from gammapy.utils.observers import observatory_locations
        from gammapy.utils.coordinates import FoVICRSFrame

        location = observatory_locations.get("ctao_north")
        obstime = Time("2025-01-01T00:00:00")
        origin = SkyCoord(85.63333333, 20.01444444, unit="deg", frame="icrs")

        fov_frame = FoVICRSFrame(origin=origin)

        crab = SkyCoord(83.63333333, 22.01444444, unit="deg", frame="icrs")
        crab.transform_to(fov_frame)
    """

    frame_specific_representation_info = {
        UnitSphericalRepresentation: [
            RepresentationMapping("lon", "fov_lon"),
            RepresentationMapping("lat", "fov_lat"),
        ]
    }
    default_representation = UnitSphericalRepresentation

    origin = CoordinateAttribute(default=None, frame=ICRS)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # make sure telescope coordinate is in range [-180°, 180°]
        if isinstance(self._data, UnitSphericalRepresentation):
            self._data.lon.wrap_angle = Angle(180, unit=u.deg)


@frame_transform_graph.transform(FunctionTransform, FoVICRSFrame, FoVICRSFrame)
def fov_icrs_to_fov_icrs(from_fov_icrs_coord, to_fov_icrs_frame):
    """Transform between two `FoVFrame`."""
    intermediate_from = from_fov_icrs_coord.transform_to(from_fov_icrs_coord.origin)
    intermediate_to = intermediate_from.transform_to(to_fov_icrs_frame.origin)
    return intermediate_to.transform_to(to_fov_icrs_frame)


@frame_transform_graph.transform(DynamicMatrixTransform, ICRS, FoVICRSFrame)
def icrs_to_fov_icrs(icrs_coord, fov_icrs_frame):
    """Convert a reference coordinate to a sky offset frame."""
    # Define rotation matrices along the position angle vector, and
    # relative to the origin.
    origin = fov_icrs_frame.origin.represent_as(UnitSphericalRepresentation)
    mat1 = rotation_matrix(-origin.lat, "y")
    mat2 = rotation_matrix(origin.lon, "z")

    return reflect_lon_matrix @ mat1 @ mat2


@frame_transform_graph.transform(DynamicMatrixTransform, FoVICRSFrame, ICRS)
def fov_icrs_to_icrs(fov_icrs_coord, icrs_frame):
    """Convert an sky offset frame coordinate to the reference frame."""
    # use the forward transform, but just invert it
    mat = icrs_to_fov_icrs(icrs_frame, fov_icrs_coord)
    return matrix_transpose(mat)


def fov_to_sky(lon, lat, lon_pnt, lat_pnt):
    """Transform field-of-view coordinates to sky coordinates.

    Parameters
    ----------
    lon, lat : `~astropy.units.Quantity`
        Field-of-view coordinate to be transformed.
    lon_pnt, lat_pnt : `~astropy.units.Quantity`
        Coordinate specifying the pointing position.
        (i.e. the center of the field of view.)

    Returns
    -------
    lon_t, lat_t : `~astropy.units.Quantity`
        Transformed sky coordinate.
    """
    # Create a frame that is centered on the pointing position
    center = SkyCoord(lon_pnt, lat_pnt)
    fov_frame = SkyOffsetFrame(origin=center)

    # Define coordinate to be transformed.
    # Need to switch the sign of the longitude angle here
    # because this axis is reversed in our definition of the FoV-system
    target_fov = SkyCoord(-lon, lat, frame=fov_frame)

    # Transform into celestial system (need not be ICRS)
    target_sky = target_fov.icrs

    return target_sky.ra, target_sky.dec


def sky_to_fov(lon, lat, lon_pnt, lat_pnt):
    """Transform sky coordinates to field-of-view coordinates.

    Parameters
    ----------
    lon, lat : `~astropy.units.Quantity`
        Sky coordinate to be transformed.
    lon_pnt, lat_pnt : `~astropy.units.Quantity`
        Coordinate specifying the pointing position.
        (i.e. the center of the field of view.)

    Returns
    -------
    lon_t, lat_t : `~astropy.units.Quantity`
        Transformed field-of-view coordinate.
    """
    # Create a frame that is centered on the pointing position
    center = SkyCoord(lon_pnt, lat_pnt)
    fov_frame = SkyOffsetFrame(origin=center)

    # Define coordinate to be transformed.
    target_sky = SkyCoord(lon, lat)

    # Transform into FoV-system
    target_fov = target_sky.transform_to(fov_frame)

    # Switch sign of longitude angle since this axis is
    # reversed in our definition of the FoV-system
    return -target_fov.lon, target_fov.lat



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from numba import jit

global TRUNCATION_VALUE  # compile time constant
TRUNCATION_VALUE = 1e-25


@jit("f8(f8[:],f8[:],f8[:])", nopython=True, nogil=True, cache=True)
def weighted_cash_sum_jit(counts, npred, weight):
    """Cash fit statistics with weights.

    Parameters
    ----------
    counts : `~numpy.ndarray`
        Counts array.
    npred : `~numpy.ndarray`
        Predicted counts array.
    weight : `~numpy.ndarray`
        likelihood weights array.
    """
    stat_sum = 0.0
    trunc = TRUNCATION_VALUE
    logtrunc = np.log(TRUNCATION_VALUE)

    ni = counts.shape[0]
    for i in range(ni):
        npr = npred[i]
        if npr > trunc:
            lognpr = np.log(npr)
        else:
            npr = trunc
            lognpr = logtrunc

        if weight[i] > 0:
            stat_sum += weight[i] * npr
            if counts[i] > 0.0:
                stat_sum -= weight[i] * counts[i] * lognpr

    return 2 * stat_sum


@jit("f8(f8[:],f8[:])", nopython=True, nogil=True, cache=True)
def cash_sum_jit(counts, npred):
    """Sum cash fit statistics.

    Parameters
    ----------
    counts : `~numpy.ndarray`
        Counts array.
    npred : `~numpy.ndarray`
        Predicted counts array.
    """
    stat_sum = 0.0
    trunc = TRUNCATION_VALUE
    logtrunc = np.log(TRUNCATION_VALUE)

    ni = counts.shape[0]
    for i in range(ni):
        npr = npred[i]
        if npr > trunc:
            lognpr = np.log(npr)
        else:
            npr = trunc
            lognpr = logtrunc

        stat_sum += npr
        if counts[i] > 0.0:
            stat_sum -= counts[i] * lognpr

    return 2 * stat_sum


@jit("f8(f8,f8[:],f8[:],f8[:])", nopython=True, nogil=True, cache=True)
def f_cash_root_jit(x, counts, background, model):
    """Function to find root of. Described in Appendix A, Stewart (2009).

    Parameters
    ----------
    x : float
        Model amplitude.
    counts : `~numpy.ndarray`
        Count image slice, where model is defined.
    background : `~numpy.ndarray`
        Background image slice, where model is defined.
    model : `~numpy.ndarray`
        Source template (multiplied with exposure).
    """
    stat_sum = 0.0
    ni = counts.shape[0]
    for i in range(ni):
        if model[i] > 0.0:
            if counts[i] > 0.0:
                denom = x * model[i] + background[i]
                if denom != 0.0:
                    stat_sum += model[i] * (1.0 - counts[i] / denom)
            else:
                stat_sum += model[i]

    # 2 is required to maintain the correct normalization of the
    # derivative of the likelihood function. It doesn't change the result of
    # the fit.
    return 2 * stat_sum


@jit("UniTuple(f8, 3)(f8[:],f8[:],f8[:])", nopython=True, nogil=True, cache=True)
def norm_bounds_jit(counts, background, model):
    """Compute bounds for the root of `_f_cash_root_jit`.

    Parameters
    ----------
    counts : `~numpy.ndarray`
        Counts image
    background : `~numpy.ndarray`
        Background image
    model : `~numpy.ndarray`
        Source template (multiplied with exposure).
    """
    s_model = 0.0
    s_counts = 0.0
    sn_min = 1e14
    c_min = 1.0
    sn_min_total = 1e14
    ni = counts.shape[0]
    for i in range(ni):
        if counts[i] > 0.0:
            s_counts += counts[i]
            if model[i] > 0.0:
                sn = background[i] / model[i]
                if sn < sn_min:
                    sn_min = sn
                    c_min = counts[i]
        if model[i] > 0.0:
            s_model += model[i]
            sn = background[i] / model[i]
            if sn < sn_min_total:
                sn_min_total = sn
    if s_model == 0.0:
        b_min = np.nan
        b_max = np.nan
    else:
        b_min = c_min / s_model - sn_min
        b_max = s_counts / s_model - sn_min
    return b_min, b_max, -sn_min_total



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from numpy.testing import assert_allclose
from astropy.units import Quantity
from gammapy.astro.source import SNR, SNRTrueloveMcKee

t = Quantity([0, 1, 10, 100, 1000, 10000], "yr")
snr = SNR()
snr_mckee = SNRTrueloveMcKee()


def test_SNR_luminosity_tev():
    """Test SNR luminosity"""
    reference = [0, 0, 0, 0, 1.076e33, 1.076e33]
    assert_allclose(snr.luminosity_tev(t).value, reference, rtol=1e-3)


def test_SNR_radius():
    """Test SNR radius"""
    reference = [0, 3.085e16, 3.085e17, 3.085e18, 1.174e19, 2.950e19]
    assert_allclose(snr.radius(t).value, reference, rtol=1e-3)


def test_SNR_radius_inner():
    """Test SNR radius"""
    reference = (1 - 0.0914) * np.array(
        [0, 3.085e16, 3.085e17, 3.085e18, 1.174e19, 2.950e19]
    )
    assert_allclose(snr.radius_inner(t).value, reference, rtol=1e-3)


def test_SNRTrueloveMcKee_luminosity_tev():
    """Test SNR Truelov McKee luminosity"""
    reference = [0, 0, 0, 0, 1.076e33, 1.076e33]
    assert_allclose(snr_mckee.luminosity_tev(t).value, reference, rtol=1e-3)


def test_SNRTrueloveMcKee_radius():
    """Test SNR RTruelove McKee radius"""
    reference = [0, 1.953e17, 9.066e17, 4.208e18, 1.579e19, 4.117e19]
    assert_allclose(snr_mckee.radius(t).value, reference, rtol=1e-3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Built-in models in Gammapy."""

from gammapy.utils.registry import Registry
from .core import DatasetModels, Model, ModelBase, Models
from .cube import (
    FoVBackgroundModel,
    SkyModel,
    TemplateNPredModel,
    create_fermi_isotropic_diffuse_model,
)
from .prior import (
    GaussianPrior,
    GeneralizedGaussianPrior,
    Prior,
    UniformPrior,
    LogUniformPrior,
)
from .spatial import (
    ConstantFluxSpatialModel,
    ConstantSpatialModel,
    DiskSpatialModel,
    GaussianSpatialModel,
    GeneralizedGaussianSpatialModel,
    PiecewiseNormSpatialModel,
    PointSpatialModel,
    Shell2SpatialModel,
    ShellSpatialModel,
    SpatialModel,
    TemplateNDSpatialModel,
    TemplateSpatialModel,
)
from .spectral import (
    EBL_DATA_BUILTIN,
    BrokenPowerLawSpectralModel,
    CompoundSpectralModel,
    ConstantSpectralModel,
    EBLAbsorptionNormSpectralModel,
    ExpCutoffPowerLaw3FGLSpectralModel,
    ExpCutoffPowerLawNormSpectralModel,
    ExpCutoffPowerLawSpectralModel,
    GaussianSpectralModel,
    LogParabolaNormSpectralModel,
    LogParabolaSpectralModel,
    NaimaSpectralModel,
    PiecewiseNormSpectralModel,
    PowerLaw2SpectralModel,
    PowerLawNormSpectralModel,
    PowerLawSpectralModel,
    ScaleSpectralModel,
    SmoothBrokenPowerLawSpectralModel,
    SpectralModel,
    SuperExpCutoffPowerLaw3FGLSpectralModel,
    SuperExpCutoffPowerLaw4FGLDR3SpectralModel,
    SuperExpCutoffPowerLaw4FGLSpectralModel,
    TemplateNDSpectralModel,
    TemplateSpectralModel,
    integrate_spectrum,
    scale_plot_flux,
)
from .spectral_cosmic_ray import create_cosmic_ray_spectral_model
from .spectral_crab import MeyerCrabSpectralModel, create_crab_spectral_model
from .temporal import (
    ConstantTemporalModel,
    ExpDecayTemporalModel,
    GaussianTemporalModel,
    GeneralizedGaussianTemporalModel,
    LightCurveTemplateTemporalModel,
    LinearTemporalModel,
    PowerLawTemporalModel,
    SineTemporalModel,
    TemplatePhaseCurveTemporalModel,
    TemporalModel,
)
from .utils import read_hermes_cube

__all__ = [
    "BrokenPowerLawSpectralModel",
    "CompoundSpectralModel",
    "ConstantFluxSpatialModel",
    "ConstantSpatialModel",
    "ConstantSpectralModel",
    "ConstantTemporalModel",
    "create_cosmic_ray_spectral_model",
    "create_crab_spectral_model",
    "create_fermi_isotropic_diffuse_model",
    "DatasetModels",
    "DiskSpatialModel",
    "EBLAbsorptionNormSpectralModel",
    "ExpCutoffPowerLaw3FGLSpectralModel",
    "ExpCutoffPowerLawNormSpectralModel",
    "ExpCutoffPowerLawSpectralModel",
    "ExpDecayTemporalModel",
    "FoVBackgroundModel",
    "GaussianSpatialModel",
    "GaussianSpectralModel",
    "GaussianTemporalModel",
    "GeneralizedGaussianSpatialModel",
    "GeneralizedGaussianTemporalModel",
    "integrate_spectrum",
    "LightCurveTemplateTemporalModel",
    "LinearTemporalModel",
    "LogParabolaNormSpectralModel",
    "LogParabolaSpectralModel",
    "MeyerCrabSpectralModel",
    "Model",
    "Models",
    "ModelBase",
    "MODEL_REGISTRY",
    "NaimaSpectralModel",
    "PiecewiseNormSpectralModel",
    "PiecewiseNormSpatialModel",
    "PointSpatialModel",
    "PowerLaw2SpectralModel",
    "PowerLawNormSpectralModel",
    "PowerLawSpectralModel",
    "PowerLawTemporalModel",
    "Prior",
    "GaussianPrior",
    "GeneralizedGaussianPrior",
    "UniformPrior",
    "LogUniformPrior",
    "read_hermes_cube",
    "scale_plot_flux",
    "ScaleSpectralModel",
    "Shell2SpatialModel",
    "ShellSpatialModel",
    "SineTemporalModel",
    "SkyModel",
    "SmoothBrokenPowerLawSpectralModel",
    "SPATIAL_MODEL_REGISTRY",
    "SpatialModel",
    "SPECTRAL_MODEL_REGISTRY",
    "SpectralModel",
    "SuperExpCutoffPowerLaw3FGLSpectralModel",
    "SuperExpCutoffPowerLaw4FGLDR3SpectralModel",
    "SuperExpCutoffPowerLaw4FGLSpectralModel",
    "TemplatePhaseCurveTemporalModel",
    "TemplateSpatialModel",
    "TemplateSpectralModel",
    "TemplateNDSpatialModel",
    "TemplateNDSpectralModel",
    "TemplateNPredModel",
    "TEMPORAL_MODEL_REGISTRY",
    "TemporalModel",
    "EBL_DATA_BUILTIN",
]


SPATIAL_MODEL_REGISTRY = Registry(
    [
        ConstantSpatialModel,
        TemplateSpatialModel,
        TemplateNDSpatialModel,
        DiskSpatialModel,
        GaussianSpatialModel,
        GeneralizedGaussianSpatialModel,
        PiecewiseNormSpatialModel,
        PointSpatialModel,
        ShellSpatialModel,
        Shell2SpatialModel,
    ]
)
"""Registry of spatial model classes."""

SPECTRAL_MODEL_REGISTRY = Registry(
    [
        ConstantSpectralModel,
        CompoundSpectralModel,
        PowerLawSpectralModel,
        PowerLaw2SpectralModel,
        BrokenPowerLawSpectralModel,
        SmoothBrokenPowerLawSpectralModel,
        PiecewiseNormSpectralModel,
        ExpCutoffPowerLawSpectralModel,
        ExpCutoffPowerLaw3FGLSpectralModel,
        SuperExpCutoffPowerLaw3FGLSpectralModel,
        SuperExpCutoffPowerLaw4FGLDR3SpectralModel,
        SuperExpCutoffPowerLaw4FGLSpectralModel,
        LogParabolaSpectralModel,
        TemplateSpectralModel,
        TemplateNDSpectralModel,
        GaussianSpectralModel,
        EBLAbsorptionNormSpectralModel,
        NaimaSpectralModel,
        ScaleSpectralModel,
        PowerLawNormSpectralModel,
        LogParabolaNormSpectralModel,
        ExpCutoffPowerLawNormSpectralModel,
    ]
)
"""Registry of spectral model classes."""

TEMPORAL_MODEL_REGISTRY = Registry(
    [
        ConstantTemporalModel,
        LinearTemporalModel,
        LightCurveTemplateTemporalModel,
        ExpDecayTemporalModel,
        GaussianTemporalModel,
        GeneralizedGaussianTemporalModel,
        PowerLawTemporalModel,
        SineTemporalModel,
        TemplatePhaseCurveTemporalModel,
    ]
)
"""Registry of temporal models classes."""

PRIOR_REGISTRY = Registry(
    [
        UniformPrior,
        GaussianPrior,
        LogUniformPrior,
    ]
)
"""Registry of prior classes."""

MODEL_REGISTRY = Registry([SkyModel, FoVBackgroundModel, TemplateNPredModel])
"""Registry of model classes"""



"""
VERITAS with Gammapy
====================

Explore VERITAS point-like DL3 files, including event lists and IRFs and
calculate Li & Ma significance, spectra, and fluxes.


Introduction
------------

`VERITAS <https://veritas.sao.arizona.edu/>`__ (Very Energetic Radiation
Imaging Telescope Array System) is a ground-based gamma-ray instrument
operating at the Fred Lawrence Whipple Observatory (FLWO) in southern
Arizona, USA. It is an array of four 12m optical reflectors for
gamma-ray astronomy in the ~ 100 GeV to > 30 TeV energy range.

VERITAS data are private and lower level analysis is done using either
the
`Eventdisplay <https://github.com/VERITAS-Observatory/EventDisplay_v4>`__
or `VEGAS (internal access
only) <https://github.com/VERITAS-Observatory/VEGAS>`__ analysis
packages to produce DL3 files (using
`V2DL3 <https://github.com/VERITAS-Observatory/V2DL3>`__), which can be
used in Gammapy to produce high-level analysis products. A small sub-set
of archival Crab nebula data has been publicly released to accompany
this tutorial, which provides an introduction to VERITAS data analysis
using gammapy for VERITAS members and external users alike.

This notebook is only intended for use with these publicly released Crab
nebula files and the use of other sources or datasets may require
modifications to this notebook.

"""

import numpy as np
from matplotlib import pyplot as plt

import astropy.units as u

from gammapy.maps import MapAxis, WcsGeom, RegionGeom

from regions import CircleSkyRegion, PointSkyRegion
from gammapy.data import DataStore
from gammapy.modeling.models import SkyModel, LogParabolaSpectralModel
from gammapy.modeling import Fit
from gammapy.datasets import Datasets, SpectrumDataset
from gammapy.estimators import FluxPointsEstimator, LightCurveEstimator
from gammapy.makers import (
    ReflectedRegionsBackgroundMaker,
    SafeMaskMaker,
    SpectrumDatasetMaker,
    WobbleRegionsFinder,
)
from astropy.coordinates import SkyCoord
from gammapy.visualization import plot_spectrum_datasets_off_regions
from gammapy.utils.regions import extract_bright_star_regions

######################################################################
# Data exploration
# ----------------
#


######################################################################
# Load in files
# ~~~~~~~~~~~~~
#
# First, we select and load VERITAS observations of the Crab Nebula. These
# files are processed with EventDisplay, but VEGAS analysis should be
# identical apart from the integration region size, which is handled by ``RAD_MAX``.
#

data_store = DataStore.from_dir("$GAMMAPY_DATA/veritas/crab-point-like-ED")
data_store.info()


######################################################################
# We filter our data by only taking observations within :math:`5^\circ`
# of the Crab Nebula. Further details on how to filter observations can be
# found in :doc:`../../user-guide/dl3`.
#

target_position = SkyCoord(83.6333, 22.0145, unit="deg")

selected_obs_table = data_store.obs_table.select_sky_circle(target_position, 5 * u.deg)
obs_ids = selected_obs_table["OBS_ID"]

observations = data_store.get_observations(obs_id=obs_ids, required_irf="point-like")


######################################################################
# Peek the first run in the data release
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#

######################################################################
# Peek the events and their time/energy/spatial distributions.
#

observations[0].events.peek()
plt.show()

######################################################################
# Peek at the IRFs included. You should verify that
# the IRFs are filled correctly and that there are no values set to zero
# within your analysis range. We can also peek at the effective area
# (``aeff``) or energy migration matrices (``edisp``) with the ``peek()``
# method.
#

observations[0].peek(figsize=(25, 5))
plt.show()

######################################################################
# Estimate counts and significance
# --------------------------------
#


######################################################################
# Set the energy binning
# ~~~~~~~~~~~~~~~~~~~~~~
#
# The energy axis will determine the bins in which energy is calculated,
# while the true energy axis defines the binning of the energy dispersion
# matrix and the effective area. Generally, the true energy axis should be
# more finely binned than the energy axis and span a larger range of
# energies, and the energy axis should be binned to match the needs of
# spectral reconstruction.
#
# Note that if the `~gammapy.makers.SafeMaskMaker` (which we will define
# later) is set to exclude events below a given percentage of the
# effective area, it will remove the entire bin containing the energy that
# corresponds to that percentage. Additionally, spectral bins are
# determined based on the energy axis and cannot be finer or offset from
# the energy axis bin edges. See
# :ref:`Safe Data Range <safe-data-range>` for more
# information on how the safe mask maker works.
#

energy_axis = MapAxis.from_energy_bounds("0.05 TeV", "100 TeV", nbin=50)
energy_axis_true = MapAxis.from_energy_bounds(
    "0.01 TeV", "110 TeV", nbin=200, name="energy_true"
)

######################################################################
# Create an exclusion mask
# ~~~~~~~~~~~~~~~~~~~~~~~~
#
# Here, we create a spatial mask and append exclusion regions for the
# source region and stars (< 6th magnitude) contained within the ``exclusion_geom``.
# We define a star exclusion region of 0.3 deg, which should contain bright stars
# within the VERITAS optical PSF.

exclusion_geom = WcsGeom.create(
    skydir=(target_position.ra.value, target_position.dec.value),
    binsz=0.01,
    width=(4, 4),
    frame="icrs",
    proj="CAR",
)

source_exclusion_region = CircleSkyRegion(center=target_position, radius=0.35 * u.deg)
exclusion_regions = extract_bright_star_regions(exclusion_geom)
exclusion_regions.append(source_exclusion_region)

exclusion_mask = ~exclusion_geom.region_mask(exclusion_regions)

######################################################################
# Define the integration region
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Point-like DL3 files can only be analyzed using the reflected regions
# background method and for a pre-determined integration region (which is
# the :math:`\sqrt{\theta^2}` used in IRF simulations).
#
# The default values for moderate/medium cuts are determined by the DL3
# file’s ``RAD_MAX`` keyword. For VERITAS data (ED and VEGAS), ``RAD_MAX``
# is not energy dependent.
#
# Note that full-enclosure files are required to use any non-point-like
# integration region.
#

on_region = PointSkyRegion(target_position)
geom = RegionGeom.create(region=on_region, axes=[energy_axis])

######################################################################
# Define the `~gammapy.makers.SafeMaskMaker`
# ------------------------------------------
#
# The `~gammapy.makers.SafeMaskMaker` sets the boundaries of our analysis based on the
# uncertainties contained in the instrument response functions (IRFs).
#
# For VERITAS point-like analysis (both ED and VEGAS), the following
# methods are strongly recommended:
#
# * ``offset-max``: Sets the maximum radial offset from the camera center within which we accept events. This is set to the edge of the VERITAS FoV.
#
# * ``edisp-bias``: Removes events which are reconstructed with energies that have :math:`>5\%` energy bias.
#
# * ``aeff-max``: Removes events which are reconstructed to :math:`<10\%` of the maximum value of the effective area. These are important to remove for spectral analysis, since they have large uncertainties on their reconstructed energies.
#

safe_mask_maker = SafeMaskMaker(
    methods=["offset-max", "aeff-max", "edisp-bias"],
    aeff_percent=10,
    bias_percent=5,
    offset_max=1.75 * u.deg,
)


######################################################################
# Data reduction
# --------------
#
# We will now run the data reduction chain to calculate our ON and OFF
# counts. To get a significance for the whole energy range (to match VERITAS packages),
# remove the `~gammapy.makers.SafeMaskMaker` from being applied to ``dataset_on_off``.
#
# The parameters of the reflected background regions can be changed using
# the `~gammapy.makers.WobbleRegionsFinder`, which is passed as an
# argument to the
# `~gammapy.makers.ReflectedRegionsBackgroundMaker`.
#

dataset_maker = SpectrumDatasetMaker(selection=["counts", "exposure", "edisp"])
dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=energy_axis_true)

region_finder = WobbleRegionsFinder(n_off_regions=16)
bkg_maker = ReflectedRegionsBackgroundMaker(
    exclusion_mask=exclusion_mask, region_finder=region_finder
)

datasets = Datasets()

for obs_id, observation in zip(obs_ids, observations):
    dataset = dataset_maker.run(dataset_empty.copy(name=str(obs_id)), observation)
    dataset = safe_mask_maker.run(dataset, observation)
    dataset_on_off = bkg_maker.run(dataset, observation)
    datasets.append(dataset_on_off)


######################################################################
# The plot below will show your exclusion regions in black and the center of your
# background regions with coloured stars. You should check to make sure
# these regions are sensible and that none of your background regions
# overlap with your exclusion regions.
#

plt.figure()
ax = exclusion_mask.plot()
plot_spectrum_datasets_off_regions(ax=ax, datasets=datasets)
plt.show()


######################################################################
# Significance analysis results
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#


######################################################################
# Here, we display the results of the significance analysis.
# ``info_table`` can be modified with ``cumulative = False`` to display a
# table with rows that correspond to the values for each run separately.
#
# However, ``cumulative = True`` is needed to produce the combined values
# in the next cell.
#

info_table = datasets.info_table(cumulative=True)
print(info_table)

print(f"ON: {info_table['counts'][-1]}")
print(f"OFF: {info_table['counts_off'][-1]}")
print(f"Significance: {info_table['sqrt_ts'][-1]:.2f} sigma")
print(f"Alpha: {info_table['alpha'][-1]:.2f}")


######################################################################
# We can also plot the cumulative excess counts and significance over
# time. For a steady source, we generally expect excess to increase
# linearly with time and for significance to increase as
# :math:`\sqrt{\textrm{time}}`.
#

fig, (ax_excess, ax_sqrt_ts) = plt.subplots(figsize=(10, 4), ncols=2, nrows=1)
ax_excess.plot(
    info_table["livetime"].to("h"),
    info_table["excess"],
    marker="o",
)

ax_excess.set_title("Excess")
ax_excess.set_xlabel("Livetime [h]")
ax_excess.set_ylabel("Excess events")

ax_sqrt_ts.plot(
    info_table["livetime"].to("h"),
    info_table["sqrt_ts"],
    marker="o",
)

ax_sqrt_ts.set_title("Significance")
ax_sqrt_ts.set_xlabel("Livetime [h]")
ax_sqrt_ts.set_ylabel("Significance [sigma]")
plt.show()


######################################################################
# Make a spectrum
# ---------------
#

######################################################################
# Now, we’ll calculate the source spectrum. This uses a forward-folding
# approach that will assume a given spectrum and fit the counts calculated
# above to that spectrum in each energy bin specified by the
# ``energy_axis``.
#
# For this reason, it’s important that spectral model be set as closely as
# possible to the expected spectrum - for the Crab nebula, this is a
# `~gammapy.modeling.models.LogParabolaSpectralModel`.
#

spectral_model = LogParabolaSpectralModel(
    amplitude=3.75e-11 * u.Unit("cm-2 s-1 TeV-1"),
    alpha=2.45,
    beta=0.15,
    reference=1 * u.TeV,
)

model = SkyModel(spectral_model=spectral_model, name="crab")

datasets.models = [model]

fit_joint = Fit()
result_joint = fit_joint.run(datasets=datasets)


######################################################################
# The best-fit spectral parameters are shown in this table.
#

print(datasets.models.to_parameters_table())


######################################################################
# We can inspect how well our data fit the model’s predicted counts in
# each energy bin.
#

ax_spectrum, ax_residuals = datasets[0].plot_fit()
ax_spectrum.set_ylim(0.1, 40)
plt.show()


######################################################################
# We can now calculate flux points to get a spectrum by fitting the
# ``result_joint`` model’s amplitude in selected energy bands (defined by
# ``energy_edges``). We set ``selection_optional = "all"`` in
# `~gammapy.estimators.FluxPointsEstimator`, which will include a calculation for the upper
# limits in bins with a significance :math:`< 2\sigma`.
#
# In the case of a non-detection or to obtain better upper limits,
# consider expanding the scan range for the norm parameter in
# `~gammapy.estimators.FluxPointsEstimator`. See
# :doc:`/tutorials/details/estimators` for more details on how to do this.
#

fpe = FluxPointsEstimator(
    energy_edges=np.logspace(-0.7, 1.5, 12) * u.TeV,
    source="crab",
    selection_optional="all",
)
flux_points = fpe.run(datasets=datasets)


######################################################################
# Now, we can plot our flux points along with the best-fit spectral model.
#

ax = flux_points.plot()
spectral_model.plot(ax=ax, energy_bounds=(0.1, 30) * u.TeV)
spectral_model.plot_error(ax=ax, energy_bounds=(0.1, 30) * u.TeV)

plt.show()


######################################################################
# Make a lightcurve and calculate integral flux
# ---------------------------------------------
#


######################################################################
# Integral flux can be calculated by integrating the spectral model we fit
# earlier. This will be a model-dependent flux estimate, so the choice of
# spectral model should match the data as closely as possible.
#
# ``e_min`` and ``e_max`` should be adjusted depending on the analysis
# requirements. Note that the actual energy threshold will use the closest
# bin defined by the ``energy_axis`` binning.
#

e_min = 0.25 * u.TeV
e_max = 30 * u.TeV

flux, flux_errp, flux_errn = result_joint.models["crab"].spectral_model.integral_error(
    e_min, e_max
)
print(
    f"Integral flux > {e_min}: {flux.value:.2} + {flux_errp.value:.2} {flux.unit} - {flux_errn.value:.2} {flux.unit}"
)


######################################################################
# Finally, we’ll create a run-wise binned light curve. See the
# :doc:`../analysis-time/light_curve_flare` tutorial for instructions on
# how to set up sub-run binning. Here, we set our energy edges so that the
# light curve has an energy threshold of 0.25 TeV and will plot upper
# limits for time bins with significance :math:`<2 \sigma`.
#

lc_maker = LightCurveEstimator(
    energy_edges=[0.25, 30] * u.TeV, source="crab", reoptimize=False
)
lc_maker.n_sigma_ul = 2
lc_maker.selection_optional = ["ul"]
lc = lc_maker.run(datasets)


######################################################################
# We can look at our results by printing the light curve as a table (with
# each line corresponding to a light curve bin) and plotting the light
# curve.
#

fig, ax = plt.subplots()
lc.sqrt_ts_threshold_ul = 2
lc.plot(ax=ax, axis_name="time", sed_type="flux")
plt.tight_layout()

table = lc.to_table(format="lightcurve", sed_type="flux")
print(table["time_min", "time_max", "flux", "flux_err"])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .core import HpxMap
from .geom import HpxGeom
from .ndmap import HpxNDMap

__all__ = [
    "HpxGeom",
    "HpxMap",
    "HpxNDMap",
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .cmap import colormap_hess, colormap_milagro
from .datasets import plot_npred_signal, plot_spectrum_datasets_off_regions
from .heatmap import annotate_heatmap, plot_heatmap
from .panel import MapPanelPlotter
from .utils import (
    add_colorbar,
    plot_contour_line,
    plot_distribution,
    plot_map_rgb,
    plot_theta_squared_table,
)

__all__ = [
    "annotate_heatmap",
    "colormap_hess",
    "colormap_milagro",
    "MapPanelPlotter",
    "add_colorbar",
    "plot_contour_line",
    "plot_heatmap",
    "plot_map_rgb",
    "plot_spectrum_datasets_off_regions",
    "plot_theta_squared_table",
    "plot_npred_signal",
    "plot_distribution",
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord
from regions import CircleSkyRegion
import gammapy.utils.parallel as parallel
from gammapy.data import GTI
from gammapy.datasets import MapDatasetOnOff
from gammapy.estimators import FluxPoints, FluxProfileEstimator
from gammapy.estimators.points.tests.test_sed import simulate_map_dataset
from gammapy.maps import MapAxis, WcsGeom
from gammapy.modeling.models import PowerLawSpectralModel
from gammapy.utils.regions import (
    make_concentric_annulus_sky_regions,
    make_orthogonal_rectangle_sky_regions,
)
from gammapy.utils.testing import requires_data, requires_dependency


def get_simple_dataset_on_off():
    axis = MapAxis.from_energy_bounds(0.1, 10, 2, unit="TeV")
    geom = WcsGeom.create(npix=40, binsz=0.01, axes=[axis])
    dataset = MapDatasetOnOff.create(geom, name="test-on-off")
    dataset.mask_safe += True
    dataset.counts += 5
    dataset.counts_off += 1
    dataset.acceptance += 1
    dataset.acceptance_off += 1
    dataset.exposure += 1000 * u.m**2 * u.s
    dataset.gti = GTI.create([0 * u.s], [5 * u.h], reference_time="2010-01-01T00:00:00")
    return dataset


def make_boxes(wcs):
    start_line = SkyCoord(0.08, -0.16, unit="deg", frame="icrs")
    end_line = SkyCoord(359.9, 0.16, unit="deg", frame="icrs")
    return make_orthogonal_rectangle_sky_regions(
        start_line, end_line, wcs, 0.1 * u.deg, 8
    )


def make_horizontal_boxes(wcs):
    start_line = SkyCoord(0.08, 0.1, unit="deg", frame="icrs")
    end_line = SkyCoord(359.9, 0.1, unit="deg", frame="icrs")
    return make_orthogonal_rectangle_sky_regions(
        start_line, end_line, wcs, 0.1 * u.deg, 8
    )


def test_profile_content():
    mapdataset_onoff = get_simple_dataset_on_off()
    wcs = mapdataset_onoff.counts.geom.wcs
    boxes = make_horizontal_boxes(wcs)

    prof_maker = FluxProfileEstimator(
        regions=boxes,
        energy_edges=[0.1, 1, 10] * u.TeV,
        selection_optional="all",
        n_sigma=1,
        n_sigma_ul=3,
    )

    result = prof_maker.run(mapdataset_onoff)

    imp_prof = result.to_table(sed_type="flux")
    assert_allclose(imp_prof[7]["x_min"], 0.1462, atol=1e-4)
    assert_allclose(imp_prof[7]["x_ref"], 0.1575, atol=1e-4)
    assert_allclose(imp_prof[7]["counts"], [[100.0], [100.0]], atol=1e-2)
    assert_allclose(imp_prof[7]["sqrt_ts"], [7.63, 7.63], atol=1e-2)
    assert_allclose(imp_prof[0]["flux"], [8e-06, 8.0e-06], atol=1e-3)

    # TODO: npred quantities are not supported by the table serialisation format
    #  so we rely on the FluxPoints object
    npred_excess = result.npred_excess.data[7].squeeze()
    assert_allclose(npred_excess, [80.0, 80.0], rtol=1e-3)

    errn = result.npred_excess_errn.data[7].squeeze()
    assert_allclose(errn, [10.75, 10.75], atol=1e-2)

    ul = result.npred_excess_ul.data[7].squeeze()
    assert_allclose(ul, [115.171875, 115.171874], atol=1e-2)


def test_profile_optimization():
    mapdataset_onoff = get_simple_dataset_on_off()
    wcs = mapdataset_onoff.counts.geom.wcs
    boxes = make_horizontal_boxes(wcs)

    with pytest.raises(ValueError):
        FluxProfileEstimator(
            regions=boxes,
            energy_edges=[0.1, 1, 10] * u.TeV,
            selection_optional="all",
            n_sigma=1,
            n_sigma_ul=3,
            reoptimize=True,
        )


def test_radial_profile():
    dataset = get_simple_dataset_on_off()
    geom = dataset.counts.geom
    regions = make_concentric_annulus_sky_regions(
        center=geom.center_skydir,
        radius_max=0.2 * u.deg,
    )

    prof_maker = FluxProfileEstimator(
        regions,
        energy_edges=[0.1, 1, 10] * u.TeV,
        selection_optional="all",
        n_sigma_ul=3,
    )
    result = prof_maker.run(dataset)

    imp_prof = result.to_table(sed_type="flux")

    assert_allclose(imp_prof[7]["x_min"], 0.14, atol=1e-4)
    assert_allclose(imp_prof[7]["x_ref"], 0.15, atol=1e-4)
    assert_allclose(imp_prof[7]["counts"], [[980.0], [980.0]], atol=1e-2)
    assert_allclose(imp_prof[7]["sqrt_ts"], [23.886444, 23.886444], atol=1e-5)
    assert_allclose(imp_prof[0]["flux"], [8e-06, 8.0e-06], atol=1e-3)

    # TODO: npred quantities are not supported by the table serialisation format
    #  so we rely on the FluxPoints object
    npred_excess = result.npred_excess.data[7].squeeze()
    assert_allclose(npred_excess, [784.0, 784.0], rtol=1e-3)

    errn = result.npred_excess_errn.data[7].squeeze()
    assert_allclose(errn, [34.075, 34.075], rtol=2e-3)

    ul = result.npred_excess_ul.data[0].squeeze()
    assert_allclose(ul, [75.834986, 75.834986], rtol=1e-3)


def test_radial_profile_one_interval():
    dataset = get_simple_dataset_on_off()
    geom = dataset.counts.geom
    regions = make_concentric_annulus_sky_regions(
        center=geom.center_skydir,
        radius_max=0.2 * u.deg,
    )

    prof_maker = FluxProfileEstimator(
        regions,
        selection_optional="all",
        energy_edges=[0.1, 10] * u.TeV,
        n_sigma_ul=3,
        sum_over_energy_groups=True,
    )
    result = prof_maker.run(dataset)

    imp_prof = result.to_table(sed_type="flux")

    assert_allclose(imp_prof[7]["counts"], [[1960]], atol=1e-5)
    assert_allclose(imp_prof[7]["npred_excess"], [[1568.0]], rtol=1e-3)
    assert_allclose(imp_prof[7]["sqrt_ts"], [33.780533], rtol=1e-3)
    assert_allclose(imp_prof[0]["flux"], [16e-06], atol=1e-3)

    axis = result.counts.geom.axes["dataset"]
    assert axis.center == ["test-on-off"]

    errn = result.npred_excess_errn.data[7].squeeze()
    assert_allclose(errn, [48.278367], rtol=2e-3)

    ul = result.npred_excess_ul.data[0].squeeze()
    assert_allclose(ul, [134.285974], rtol=1e-3)


def test_serialisation(tmpdir):
    dataset = get_simple_dataset_on_off()
    geom = dataset.counts.geom
    regions = make_concentric_annulus_sky_regions(
        center=geom.center_skydir,
        radius_max=0.2 * u.deg,
    )

    est = FluxProfileEstimator(regions, energy_edges=[0.1, 10] * u.TeV)
    result = est.run(dataset)

    result.write(tmpdir / "profile.fits")

    profile = FluxPoints.read(
        tmpdir / "profile.fits",
        reference_model=PowerLawSpectralModel(),
    )

    assert_allclose(result.norm, profile.norm, rtol=1e-4)
    assert_allclose(result.norm_err, profile.norm_err, rtol=1e-4)
    assert_allclose(result.npred, profile.npred)
    assert_allclose(result.ts, profile.ts)
    assert_allclose(profile.gti.time_start[0].mjd, 55197.000766, rtol=1e-5)

    assert np.all(result.is_ul == profile.is_ul)


def test_regions_init():
    with pytest.raises(ValueError):
        FluxProfileEstimator(regions=[])

    region = CircleSkyRegion(center=SkyCoord("0d", "0d"), radius=0.1 * u.deg)

    with pytest.raises(ValueError):
        FluxProfileEstimator(regions=[region])


@requires_data()
def test_profile_with_model_or_mask():
    dataset = simulate_map_dataset(name="test-map-pwl")

    geom = dataset.counts.geom
    regions = make_concentric_annulus_sky_regions(
        center=geom.center_skydir,
        radius_max=0.2 * u.deg,
    )

    prof_maker = FluxProfileEstimator(
        regions,
        selection_optional="all",
        energy_edges=[0.1, 10] * u.TeV,
        n_sigma_ul=3,
        sum_over_energy_groups=True,
    )
    result = prof_maker.run(dataset)
    imp_prof = result.to_table(sed_type="flux")
    assert_allclose(imp_prof[7]["npred_excess"], [[-1.115967]], rtol=1e-3)

    dataset.models = None
    result = prof_maker.run(dataset)
    imp_prof = result.to_table(sed_type="flux")
    assert_allclose(imp_prof[7]["npred_excess"], [[112.95312]], rtol=1e-3)

    dataset.mask_fit = ~geom.region_mask([regions[7]])
    result = prof_maker.run(dataset)
    imp_prof = result.to_table(sed_type="flux")
    assert_allclose(imp_prof[7]["npred_excess"], [[0]], rtol=1e-3)


@requires_data()
def test_profile_multiprocessing():
    dataset = simulate_map_dataset(name="test-map-pwl")

    geom = dataset.counts.geom
    regions = make_concentric_annulus_sky_regions(
        center=geom.center_skydir,
        radius_max=0.2 * u.deg,
    )

    prof_maker = FluxProfileEstimator(
        regions,
        selection_optional="all",
        energy_edges=[0.1, 10] * u.TeV,
        n_sigma_ul=3,
        sum_over_energy_groups=True,
        n_jobs=2,
        parallel_backend="multiprocessing",
    )
    result = prof_maker.run(dataset)
    imp_prof = result.to_table(sed_type="flux")
    assert_allclose(imp_prof[7]["npred_excess"], [[-1.115967]], rtol=1e-3)


@requires_data()
@requires_dependency("ray")
def test_profile_multiprocessing_ray_with_manager():
    dataset = simulate_map_dataset(name="test-map-pwl")

    geom = dataset.counts.geom
    regions = make_concentric_annulus_sky_regions(
        center=geom.center_skydir,
        radius_max=0.2 * u.deg,
    )

    with parallel.multiprocessing_manager(backend="ray", pool_kwargs=dict(processes=2)):
        prof_maker = FluxProfileEstimator(
            regions,
            selection_optional="all",
            energy_edges=[0.1, 10] * u.TeV,
            n_sigma_ul=3,
            sum_over_energy_groups=True,
        )
        assert prof_maker.n_jobs == 2
        assert prof_maker.parallel_backend == "ray"
        result = prof_maker.run(dataset)
        imp_prof = result.to_table(sed_type="flux")
        assert_allclose(imp_prof[7]["npred_excess"], [[-1.115967]], rtol=1e-3)


@requires_data()
@requires_dependency("ray")
def test_profile_multiprocessing_ray():
    dataset = simulate_map_dataset(name="test-map-pwl")

    geom = dataset.counts.geom
    regions = make_concentric_annulus_sky_regions(
        center=geom.center_skydir,
        radius_max=0.2 * u.deg,
    )

    prof_maker = FluxProfileEstimator(
        regions,
        selection_optional="all",
        energy_edges=[0.1, 10] * u.TeV,
        n_sigma_ul=3,
        sum_over_energy_groups=True,
        n_jobs=2,
        parallel_backend="ray",
    )
    result = prof_maker.run(dataset)
    imp_prof = result.to_table(sed_type="flux")
    assert_allclose(imp_prof[7]["npred_excess"], [[-1.115967]], rtol=1e-3)
